@
@ Little test program for the ARM
@
@ It sets up a few exception vector handlers, each of which
@ write an exception code to the "ExceptionType" variable and loops
@ indefinitely.
@
@ If this program is uploaded & run, the ARM will first execute the
@ ResetHandler, and iterate in the infinite loop there.
@ The SH4 can check which handler has been entered by peeking at the
@ ExceptionType variable.
@

		.ARM
		.GLOBL	_start

	
_start:
		b	ResetHandler			@ Reset vector
		b	UndefinedInstructionHandler	@ Undefined instruction vector
		b	SoftwareIntHandler		@ Software interrupt vector
		b	AbortPrefetchHandler		@ Abort (Prefetch) vector
		b	AbortDataHandler		@ Abort (Data) vector
		.LONG	0				@ Reserved
		b	IRQHandler			@ Normal IRQ vector
		b	FIQHandler			@ Fast IRQ vector

ExceptionType:	.LONG	0


ResetHandler:					@ This code is run when the ARM is booted/reseted
		mov	r0,#0x80000000
		str	r0,ExceptionType
		b	ResetHandler
	
UndefinedInstructionHandler:
		mov	r0,#0x81000000
		str	r0,ExceptionType
		b	UndefinedInstructionHandler

SoftwareIntHandler:
		mov	r0,#0x82000000
		str	r0,ExceptionType
		b	SoftwareIntHandler

AbortPrefetchHandler:
		mov	r0,#0x83000000
		str	r0,ExceptionType
		b	AbortPrefetchHandler

AbortDataHandler:
		mov	r0,#0x84000000
		str	r0,ExceptionType
		b	AbortDataHandler

IRQHandler:
		mov	r0,#0x85000000
		str	r0,ExceptionType
		b	IRQHandler

FIQHandler:
		mov	r0,#0x86000000
		str	r0,ExceptionType
		b	FIQHandler

