
#include <stdio.h>


#define BLOCKSIZE 1024
#define BYTESPERROW 20

FILE* infile;
FILE* outfile;

void cleanup(void)
{
  if (infile && infile != stdin)
    fclose(infile);
  if (outfile && outfile != stdout)
    fclose(outfile);
}

int main(int argc, char **argv)
{
  int len = 0;
  int eof = 0;

  infile = stdin;
  outfile = stdout;

  if (argc < 2 || argc > 4)
    {
      printf("Binary to SH .s converter\n");
      printf("By Mikael Kalms 2001-01-20\n\n");
      printf("Usage: bin2sh-as <label name> [input file] [output file]\n\n");
      printf("If input file is omitted, standard input will be used\n");
      printf("If output file is omitted, standard output will be used\n\n");
      return 0;
    }

  if (argc >= 3)
    {
      if (!(infile = fopen(argv[2], "rb")))
	{
	  fprintf(stderr, "Unable to open input file \"%s\"\n", argv[2]);
	  cleanup();
	  return 1;
	}
    }
  if (argc >= 4)
    {
      if (!(outfile = fopen(argv[3], "wt")))
	{
	  fprintf(stderr, "Unable to open output file \"%s\"\n", argv[3]);
	  cleanup();
	  return 1;
	}
    }

  if (!fprintf(outfile, ".align 5\n.globl %s\n.globl %s_end\n%s:\n", argv[1], argv[1], argv[1]))
    {
	  fprintf(stderr, "Error while writing to output file \"%s\": %s\n", argv[3], strerror(ferror(outfile)));
	  cleanup();
	  return 1;
    }

  while (!eof)
    {
      unsigned char cbuf[BLOCKSIZE];
      int inbytes;
      int outpos = 0;

      if ((inbytes = fread(cbuf, 1, BLOCKSIZE, infile)) < BLOCKSIZE)
	{
	  if (ferror(infile))
	    {
	      fprintf(stderr, "Error while reading from input file \"%s\": %s\n", argv[2], strerror(ferror(infile)));
	      cleanup();
	      return 1;
	    }
	  else
	    eof = 1;
	}

      while (inbytes)
	{
	  fprintf(outfile, "%s%d%s", len ? "," : ".byte ", cbuf[outpos], (len == BYTESPERROW - 1) ? "\n" : "");
	  if (++len == BYTESPERROW)
	    len = 0;
	  outpos++;
	  inbytes--;
	}

      if (ferror(outfile))
	{
	  fprintf(stderr, "Error while writing to output file \"%s\": %s\n", argv[3], strerror(ferror(outfile)));
	  cleanup();
	  return 1;
	}
    }

  fprintf(outfile, "\n%s_end:\n", argv[1]);
  if (ferror(outfile))
    {
      fprintf(stderr, "Error while writing to output file \"%s\": %s\n", argv[3], strerror(ferror(outfile)));
      cleanup();
      return 1;
    }

  cleanup();
  return 0;
}



