!
! TRAPA exceptions:
!
! 0x00	Dump regs
! 0x01	Dump regs + reset
! 0x02	Print string				r4: string
! 0x03	Print string + dump regs + reset	r4: string


		.INCLUDE "SH4_Regs.i"

		.SECTION .text


!---------------------------------------------------------------------
! void SetupExceptionHandler(void)
!
! Sets up VBR to own exception handlers, and setup SR bits
! (set RB = 0, BL = 0)
!	

		.GLOBL	_SetupExceptionHandler
		.ALIGN	5
_SetupExceptionHandler:
		sts.l	pr,@-r15

		mova	HandleException_100,r0
		mov	#1,r5
		bsr	_SetupExceptionHandler_buildJumpVector
		mov	r0,r4

		mova	HandleException_400,r0
		mov	#4,r5
		bsr	_SetupExceptionHandler_buildJumpVector
		mov	r0,r4

		mova	HandleException_600,r0
		mov	#6,r5
		bsr	_SetupExceptionHandler_buildJumpVector
		mov	r0,r4

		mov.l	_SetupExceptionHandler_WritebackOC,r0
		jsr	@r0
		nop

		mov.l	_SetupExceptionHandler_ExceptionPtrTable,r1
		mova	HandleException_defaultHandler,r0
		mov	#0x800/32,r2
_SetupExceptionHandler_resetHandler:
		dt	r2
		bf/s	_SetupExceptionHandler_resetHandler
		mov.l	r0,@-r1

		mova	HandleException_trapaHandler,r0
		mov	r0,r2
		mov	#0x160/32*4,r0
		mov.l	r2,@(r0,r1)

		mov.l	_SetupExceptionHandler_VBRCodeStart,r0
		ldc	r0,vbr
		stc	sr,r1
		mov	#0xcf,r0
		shll16	r0
		shll8	r0
		or	#0xff,r0
		and	r1,r0
		ldc	r0,sr

		lds.l	@r15+,pr
		rts
		nop

_SetupExceptionHandler_buildJumpVector:
		mov.l	_SetupExceptionHandler_VBRCodeStart,r3
		shll8	r5
		mov	#0x0009,r0
		mov.l	_SetupExceptionHandler_MovJmp,r2
		add	r5,r3
		mov.w	r0,@(0,r3)
		mov.w	r0,@(2,r3)
		mov.l	r2,@(4,r3)
		mov.w	r0,@(8,r3)
		mov.w	r0,@(10,r3)
		rts
		mov.l	r4,@(12,r3)

					.ALIGN	3
_SetupExceptionHandler_VBRCodeStart:	.LONG	VBRCodeStart
_SetupExceptionHandler_MovJmp:		.LONG	0x402bd001
_SetupExceptionHandler_WritebackOC:	.LONG	_WritebackOC
_SetupExceptionHandler_ExceptionPtrTable: .LONG	ExceptionPtrTable+(0x800/32*4)



!---------------------------------------------------------------------
! void DisableExceptionHandler(void)
!
! Disables exception handling, and restores SR bits
! (set BL = 1, and set IMASK = 0xf)
!	

		.GLOBL	_DisableExceptionHandler
		.ALIGN	5
_DisableExceptionHandler:
		stc	sr,r1
		mov	#0x10,r0
		shll16	r0
		shll8	r0
		or	#0xf0,r0
		or	r1,r0
		ldc	r0,sr
		rts
		nop



!---------------------------------------------------------------------
! int oldIMASK = SetIMASK(int newIMASK)
!
! Sets SR.IMASK to the given value, and returns the old value
!
! in	r4	int	new IMASK (0 .. 0xf)
! out	r0	int	old IMASK (0 .. 0xf)
!

		.GLOBL	_SetIMASK
		.ALIGN	5
_SetIMASK:
		stc	sr,r0
		shll2	r4
		mov	#0xc3,r1
		shll2	r1
		shll2	r4
		and	r0,r1
		or	r4,r1
		shlr2	r0
		ldc	r1,sr
		shlr2	r0
		rts
		and	#0xf,r0



!---------------------------------------------------------------------
! void SetExceptionEventHandler(int eventcode, void* handler)
!
! Sets one exception handler vector to point to user code
!
! in	r4	int	event code (0 .. 0x800)
!	r5	void*	handler routine
!	

		.GLOBL	_SetExceptionEventHandler
		.ALIGN	5
_SetExceptionEventHandler:
		shlr2	r4
		mov	#0xfc,r1
		mov.l	_SetExceptionEventHandler_ExceptionPtrTable,r0
		shlr	r4
		and	r1,r4
		mov.l	r5,@(r0,r4)
		rts
		nop

						.ALIGN	2
_SetExceptionHandler_WritebackOC:		.LONG	_WritebackOC
_SetExceptionEventHandler_ExceptionPtrTable:	.LONG	ExceptionPtrTable



!---------------------------------------------------------------------
! void SetupExceptionHandler(void)
!
! Cause an "Illlegal data address" exception
!	

		.GLOBL	_ForceException
		.ALIGN	5
_ForceException:
		mov	#0x8c,r0
		shll16	r0
		shll8	r0
		add	#2,r0
		mov.l	@r0,r1
		rts
		nop



!--------------------------------------------
! Exception handling code

HandleException_100:
		mov	#1,r4
		shll8	r4
		mov	#0xff,r0
		shll16	r0
		shll8	r0
		bra	HandleException
		mov.l	@(EXPEVT&0xff,r0),r5

HandleException_400:
		mov	#4,r4
		shll8	r4
		mov	#0xff,r0
		shll16	r0
		shll8	r0
		bra	HandleException
		mov.l	@(EXPEVT&0xff,r0),r5

HandleException_600:
		mov	#6,r4
		shll8	r4
		mov	#0xff,r0
		shll16	r0
		shll8	r0
		bra	HandleException
		mov.l	@(INTEVT&0xff,r0),r5

HandleException:
		mov	r5,r1
		shlr2	r1
		mov.l	HandleException_ExceptionPtrTable,r0
		shlr	r1
		mov	#0xfc,r2
		and	r2,r1
		mov.l	@(r0,r1),r1

		mova	ExceptionStack,r0
		mov	r0,r15
		nop	
		sts.l	pr,@-r15
		jsr	@r1
		nop
		lds.l	@r15+,pr

		stc	sgr,r15
		clrs
		clrs
		clrs
		clrs
		clrs
		clrs
		rte			! 19+ cycles of margin before RTE please
		nop

					.ALIGN	2
HandleException_ExceptionPtrTable:	.LONG	ExceptionPtrTable

HandleException_trapaHandler:
		mov	#0xff,r0
		shll16	r0
		shll8	r0
		mov.l	@(0x20,r0),r0
		shlr2	r0
		cmp/eq	#0,r0
		bt	HandleException_trapaHandler_0
		cmp/eq	#1,r0
		bt	HandleException_defaultHandler
		cmp/eq	#2,r0
		bt	HandleException_trapaHandler_2
		cmp/eq	#3,r0
		bt	HandleException_trapaHandler_3
		bra	HandleException_defaultHandler
		nop

HandleException_trapaHandler_0:
		sts.l	pr,@-r15
		bsr	HandleException_dumpRegs
		nop
		lds.l	@r15+,pr
		rts
		nop

HandleException_trapaHandler_2:
		mov.l	r5,@-r15
		mov.l	r4,@-r15
		sts.l	pr,@-r15
		mov.l	HandleException_SerialWriteString,r0
		jsr	@r0
		stc	r4_bank,r4
		lds.l	@r15+,pr
		mov.l	@r15+,r4
		mov.l	@r15+,r5
		rts
		nop

HandleException_trapaHandler_3:
		mov.l	r5,@-r15
		mov.l	r4,@-r15
		sts.l	pr,@-r15
		mov.l	HandleException_SerialWriteString,r0
		jsr	@r0
		stc	r4_bank,r4
		lds.l	@r15+,pr
		mov.l	@r15+,r4
		mov.l	@r15+,r5

		bra	HandleException_defaultHandler
		nop

HandleException_defaultHandler:
		bsr	HandleException_dumpRegs
		nop

		mov	#-1,r0		! Force warm boot
		ldc	r0,sr
		mov.l	@r0,r0
		nop
		nop

		
HandleException_dumpRegs:
		sts.l	pr,@-r15
		mov.l	r5,@-r15
		mov.l	r4,@-r15

		mova	HandleException_OldRn+16*4,r0
		stc.l	sgr,@-r0
		mov.l	r14,@-r0
		mov.l	r13,@-r0
		mov.l	r12,@-r0
		mov.l	r11,@-r0
		mov.l	r10,@-r0
		mov.l	r9,@-r0
		mov.l	r8,@-r0
		stc.l	r7_bank,@-r0
		stc.l	r6_bank,@-r0
		stc.l	r5_bank,@-r0
		stc.l	r4_bank,@-r0
		stc.l	r3_bank,@-r0
		stc.l	r2_bank,@-r0
		stc.l	r1_bank,@-r0
		stc.l	r0_bank,@-r0

		mov.l	HandleException_SerialWriteString,r8
		mov.l	HandleException_SerialWriteHexLong,r9
		mov.l	HandleException_SerialWriteChar,r10

		mova	ExceptionString1,r0
		jsr	@r8
		mov	r0,r4

		jsr	@r9
		mov.l	@r15+,r4

		mova	ExceptionString2,r0
		jsr	@r8
		mov	r0,r4

		jsr	@r9
		mov.l	@r15+,r4

		mova	ExceptionString3,r0
		jsr	@r8
		mov	r0,r4

		mov	#0xff,r0
		shll16	r0
		shll8	r0
		jsr	@r9
		mov.l	@(0xc,r0),r4

		mova	HandleException_OldRn,r0
		mov	r0,r12
		mov	#4,r14
HandleException_RnLoop2:

		mova	ExceptionString4,r0
		jsr	@r8
		mov	r0,r4

		mov	#4,r13
HandleException_RnLoop1:

		jsr	@r9
		mov.l	@r12+,r4

		jsr	@r10
		mov	#' ',r4

		dt	r13
		bf	HandleException_RnLoop1

		dt	r14
		bf	HandleException_RnLoop2
	
		mova	ExceptionString5,r0
		jsr	@r8
		mov	r0,r4

		jsr	@r9
		stc	spc,r4

		mova	ExceptionString6,r0
		jsr	@r8
		mov	r0,r4

		jsr	@r9
		stc	ssr,r4

		mova	ExceptionString7,r0
		jsr	@r8
		mov	r0,r4

		mova	HandleException_OldRn+8*4,r0
		mov.l	@r0+,r8
		mov.l	@r0+,r9
		mov.l	@r0+,r10
		mov.l	@r0+,r11
		mov.l	@r0+,r12
		mov.l	@r0+,r13
		mov.l	@r0+,r14
	
		lds.l	@r15+,pr
		rts
		nop

					.ALIGN	2
HandleException_SerialWriteString:	.LONG	_SerialWriteString
HandleException_SerialWriteHexLong:	.LONG	_SerialWriteHexLong
HandleException_SerialWriteChar:	.LONG	_SerialWriteChar
HandleException_OldRn:			.SPACE	16*4
	
			.ALIGN	2
ExceptionString1:	.ASCIZ	"*****EXCEPTION*****\r\nType:   "
			.ALIGN	2
ExceptionString2:	.ASCIZ	"\r\nEvent:  "
			.ALIGN	2
ExceptionString3:	.ASCIZ	"\r\nTEA:    "
			.ALIGN	2
ExceptionString4:	.ASCIZ	"\r\nRn: "
			.ALIGN	2
ExceptionString5:	.ASCIZ	"\r\nPC: "
			.ALIGN	2
ExceptionString6:	.ASCIZ	"\r\nSR: "
			.ALIGN	2
ExceptionString7:	.ASCIZ	"\r\n"

		.SECTION .bss
	
			.ALIGN	5
ExceptionPtrTable:	.SPACE	0x800/32*4

			.ALIGN	5
			.SPACE	256
ExceptionStack:		.SPACE	16

			.ALIGN	5
VBRCodeStart:		.SPACE	0x620

