!
! NOTE: The VGA and NTSC support in here has only been tested
!	very rudimentarily. It may or may not work.
!

		.INCLUDE "PVR2_Regs.i"
		.INCLUDE "SH4_Regs.i"
		.INCLUDE "Screen.i"

		.SECTION .text


!-----------------------------------------------------------------------------
! void scWaitVBL(void)
!
! Idles until raster beam has reached top of screen
!

		.GLOBL	_scWaitVBL
		.ALIGN	5
_scWaitVBL:
		mov.l	_scWaitVBL_PVR2_BeamPos,r1

_scWaitVBL_sync1:
		mov.l	@r1,r0
		shlr8	r0
		tst	#0x20,r0
		bf	_scWaitVBL_sync1

_scWaitVBL_sync2:
		mov.l	@r1,r0
		shlr8	r0
		tst	#0x20,r0
		bt	_scWaitVBL_sync2

		rts
		nop

				.ALIGN	2
_scWaitVBL_PVR2_BeamPos:	.long	PVR2REGS+BEAMPOS*4



!------------------------------------------------------------------------
! void scInit(void)
!
! Initializes screen subsystem - will leave system with the original
! SEGA logo showing
!
		.GLOBL	_scInit
		.ALIGN	5
_scInit:
		sts.l	pr,@-r15

		bsr	scGetDisplayType
		nop
		mov.l	_scInit_Screen,r1
		mov.b	r0,@(Screen_DisplayType,r1)

		add	#-8,r15
		tst	#DISPLAY_DEFAULT_60HZ,r0
		movt	r1
		add	#-1,r1
		mov.w	_scInit_Width,r4
		mov.w	_scInit_Height,r5
		mov	#SCREEN_HIRES,r6
		mov.l	r1,@(0,r15)
		mov	#SCREEN_ARGB8888,r0
		mov	#SCREEN_HIRES,r7
		bsr	_scSetMode
		mov.l	r0,@(4,r15)
		add	#8,r15

		mov.l	_scInit_Base,r4
		mov.w	_scInit_Width,r5
		bsr	_scSetBuffer
		shll2	r5

		mov	#0,r4
		bsr	_scSetCenter
		mov	#0,r5
	
		bsr	_scEnable
		mov	#1,r4

		lds.l	@r15+,pr
		rts
		nop

		.ALIGN	4
_scInit_Width:	.word	640
_scInit_Height:	.word	480
_scInit_Screen:	.long	_Screen
_scInit_Base:	.long	0x200000



!------------------------------------------------------------------------
! void scEnable(bool enable)
!
! Turns the display on/off
!
! in	r4		bool	enable
!

		.GLOBL	_scEnable
		.ALIGN	5
_scEnable:
		mova	_Screen,r0
		tst	r4,r4
		mov.l	_scEnable_PVR2Regs,r2
		mov	r0,r1
		bt	_scEnable_disable

		mov.b	@(Screen_Flags,r1),r0
		or	#SCREEN_FLAGS_ENABLE,r0
		mov.b	r0,@(Screen_Flags,r1)
		mov	#BITMAPTYPE*4,r0
		mov	#1,r3
		mov.l	@(r0,r2),r1
		or	r3,r1
		mov.l	r1,@(r0,r2)
		mov	#BITMAPTYPE2,r0
		shll2	r0
		mov	#~8,r3
		mov.l	@(r0,r2),r1
		and	r3,r1
		rts
		mov.l	r1,@(r0,r2)

_scEnable_disable:
		mov.b	@(Screen_Flags,r1),r0
		and	#~SCREEN_FLAGS_ENABLE,r0
		mov.b	r0,@(Screen_Flags,r1)
		mov	#BITMAPTYPE*4,r0
		mov	#~1,r3
		mov.l	@(r0,r2),r1
		and	r3,r1
		mov.l	r1,@(r0,r2)
		mov	#BITMAPTYPE2,r0
		shll2	r0
		mov	#8,r3
		mov.l	@(r0,r2),r1
		or	r3,r1
		rts
		mov.l	r1,@(r0,r2)

			.ALIGN	2
_scEnable_PVR2Regs:	.long	PVR2REGS

	

!------------------------------------------------------------------------
! void scSetMode(int width, int height, int horizClk, int vertClk, int freq, int bpp)
!
! Sets up display mode with the given parameters
!
! in	r4		int	width (in pixels) [must be even]
!	r5		int	height (in pixels) [must be even]
!	r6		int	horizClk (0 = lowres, 1 = highres)
!	r7		int	vertClk (0 = lowres, 1 = highres)
!	@(0,r15)	int	freq (0 = 50Hz, 1 = 60Hz)
!	@(4,r15)	int	bpp (0..3, BITMAP_* equates)
!

		.GLOBL	_scSetMode
		.ALIGN	5
_scSetMode:
		mova	_Screen,r0
		mov	r0,r3
		sts.l	pr,@-r15
		mov	#-2,r1
		mov.l	r6,@(Screen_HorizClock,r3)
		and	r1,r4
		mov.l	r7,@(Screen_VertClock,r3)
		and	r1,r5
		mov.l	r4,@(Screen_Width,r3)
		mov.l	@(4,r15),r1
		mov.l	r1,@(Screen_RefreshFreq,r3)
		mov.l	@(8,r15),r1
		mov.l	r5,@(Screen_Height,r3)
		bsr	scUpdateMode
		mov.l	r1,@(Screen_Format,r3)
		lds.l	@r15+,pr

		rts
		nop



!------------------------------------------------------------------------
! void scSetBuffer(int base, int pitch)
!
! Sets up the buffer which is to be shown in the bitmap window
!
! in	r4		int	base
!	r5		int	pitch
!

		.GLOBL	_scSetBuffer
		.ALIGN	5
_scSetBuffer:
		mova	_Screen,r0
		mov	r0,r3
		sts.l	pr,@-r15
		mov.l	r4,@(Screen_Base,r3)
		bsr	scUpdateBuffer
		mov.l	r5,@(Screen_Pitch,r3)
		lds.l	@r15+,pr

		rts
		nop



!------------------------------------------------------------------------
! void scSetCenter(int xcenter, int ycenter)
!
! Adjusts center of bitmap window; (0,0) is default, and screen center
!
! in	r4	int	xcenter (in highres pixels)
!	r5	int	ycenter (in highres pixels)
!

		.GLOBL	_scSetCenter
		.ALIGN	5
_scSetCenter:
		mova	_Screen,r0
		mov	r0,r3
		sts.l	pr,@-r15
		mov.l	r4,@(Screen_XCenter,r3)
		bsr	scUpdateCenter
		mov.l	r5,@(Screen_YCenter,r3)
		lds.l	@r15+,pr

		rts
		nop



!------------------------------------------------------------------------
! void scSetBorderColor(int color)
!
! Sets the colour of the border window
!
! in	r4	int	colour (ARGB8888 format)
!

		.GLOBL	_scSetBorderColor
		.ALIGN	5
_scSetBorderColor:
		mov.l	_scSetBorderColor_PVR2_BorderColor,r0
		rts
		mov.l	r4,@r0

					.ALIGN	2
_scSetBorderColor_PVR2_BorderColor:	.long	PVR2REGS+BORDERCOLOR*4



!-----------------------------------------------------------------------------
! int scGetDisplayType(void)
!
! Retreives display broadcast/cable/system type, and default refresh rate
! In most cases, it is enough to just use the display system type information
!
! out	r0	int	displayType
!

		.ALIGN	5
scGetDisplayType:
		sts.l	pr,@-r15
		mov.l	@(0xb0,gbr),r0			! Get SysConfigBlock
		jsr	@r0
		mov	#3,r7

		add	#0xc,r0
		mov.b	@r0,r1				! Get display broadcast type
		mov	#-'0',r2
		add	r2,r1
		mov	#3,r3
		cmp/hi	r3,r1
		bf	scGetDisplayType_broadcastOk	! If invalid, default to NTSC
		mov	#0,r1
scGetDisplayType_broadcastOk:
	
		mov.l	scGetDisplayType_CCN_PDTRA,r0
		shll2	r1
		mov.w	@r0,r2				! Get display cable type
		shlr8	r2
		and	r3,r2

		mova	scGetDisplayType_SystemTypes,r0
		or	r2,r1
		mov.b	@(r0,r1),r0			! Get display system type (PALN -> PAL, PALM -> NTSC)
		shll2	r1

		lds.l	@r15+,pr
		rts
		or	r1,r0

				.ALIGN	5
scGetDisplayType_CCN_PDTRA:	.long	0xff800030
scGetDisplayType_SystemTypes:
		.byte	DISPLAY_SYSTEM_VGA | DISPLAY_DEFAULT_60HZ, DISPLAY_SYSTEM_UNKNOWN	! NTSC broadcast
		.byte	DISPLAY_SYSTEM_NTSC | DISPLAY_DEFAULT_60HZ, DISPLAY_SYSTEM_NTSC | DISPLAY_DEFAULT_60HZ
	
		.byte	DISPLAY_SYSTEM_VGA | DISPLAY_DEFAULT_60HZ, DISPLAY_SYSTEM_UNKNOWN	! PAL broadcast
		.byte	DISPLAY_SYSTEM_PAL | DISPLAY_DEFAULT_50HZ, DISPLAY_SYSTEM_PAL | DISPLAY_DEFAULT_50HZ

		.byte	DISPLAY_SYSTEM_VGA | DISPLAY_DEFAULT_60HZ, DISPLAY_SYSTEM_UNKNOWN	! PALM broadcast
		.byte	DISPLAY_SYSTEM_NTSC | DISPLAY_DEFAULT_60HZ, DISPLAY_SYSTEM_NTSC | DISPLAY_DEFAULT_60HZ

		.byte	DISPLAY_SYSTEM_VGA | DISPLAY_DEFAULT_60HZ, DISPLAY_SYSTEM_UNKNOWN	! PALN broadcast
		.byte	DISPLAY_SYSTEM_PAL | DISPLAY_DEFAULT_50HZ, DISPLAY_SYSTEM_PAL | DISPLAY_DEFAULT_50HZ



!------------------------------------------------------------------------
! void scUpdateMode(void)
!
! Updates the following registers:
!
! BITMAPTYPE, BITMAPTYPE2, IMAGECONTROL, FRAMETOTAL, DIWSIZE(height),
! BORDERWINDOWX, BORDERWINDOWY
! Calls scUpdateBuffer, scUpdateWindow
!

		.ALIGN	5
scUpdateMode:
		mov.l	r8,@-r15
		sts.l	pr,@-r15

		mova	_Screen,r0
		mov	r0,r6
		mov.b	@(Screen_DisplayType,r6),r0

		and	#DISPLAY_SYSTEM_MASK,r0
		cmp/eq	#DISPLAY_SYSTEM_VGA,r0
		mova	scUpdateMode_Regs_Tables,r0
		bt	scUpdateMode_VGA

		mov.l	@(Screen_RefreshFreq,r6),r1			! TV may support 50Hz or 60Hz
		bra	scUpdateMode_common
		shll2	r1

scUpdateMode_VGA:
		bra	scUpdateMode_common
		mov	#2*4,r1						! VGA only supports 60Hz

		.ALIGN	4
scUpdateMode_Regs_Tables:
		.long	scUpdateMode_Regs_15kHz_50Hz
		.long	scUpdateMode_Regs_15kHz_60Hz
		.long	scUpdateMode_Regs_31kHz_60Hz

scUpdateMode_common:
		mov.l	@(r0,r1),r8

		mov.l	@(DisplayModeInfo_DisplayOriginX,r8),r0
		mov.l	@(DisplayModeInfo_DisplayOriginY,r8),r1
		mov.l	r0,@(Screen_DisplayOriginX,r6)
		mov.l	r1,@(Screen_DisplayOriginY,r6)

		mov.l	@(DisplayModeInfo_HorizFreq,r8),r2
		tst	r2,r2
		mov.l	@(Screen_VertClock,r6),r1
		movt	r2
		mov.l	@(Screen_HorizClock,r6),r0
		tst	r1,r1
		addc	r2,r2
		tst	r0,r0
		addc	r2,r2
		mova	scUpdateMode_Table0,r0
		mov.b	@(r0,r2),r0
		and	#~SCREEN_FLAGS_ENABLE,r0
		mov.l	@(Screen_RefreshFreq,r6),r2
		mov	r0,r1
		mov.b	@(Screen_Flags,r6),r0
		tst	r2,r2
		and	#SCREEN_FLAGS_ENABLE,r0
		movt	r2
		or	r1,r0
		add	#-1,r2
		mov	#SCREEN_FLAGS_60HZ,r3
		and	r3,r2
		or	r2,r0
		mov.b	r0,@(Screen_Flags,r6)

		mov.l	scUpdateMode_PVR2Regs,r5

		mov.l	@(Screen_Format,r6),r1
		shll2	r1
		mov	#1,r2
		or	r2,r1
		mov.b	@(Screen_Flags,r6),r0
		tst	#SCREEN_FLAGS_31KHZ,r0
		mov	#23,r3
		subc	r2,r2
		add	#1,r2
		shld	r3,r2
		or	r2,r1
		tst	#SCREEN_FLAGS_DOUBLESCAN,r0
		subc	r2,r2
		add	#1,r2
		add	r2,r2
		or	r2,r1
		tst	#SCREEN_FLAGS_ENABLE,r0
		subc	r2,r2
		add	#1,r2
		or	r2,r1
		mov	#BITMAPTYPE,r0
		shll2	r0
		mov.l	r1,@(r0,r5)

		mov	#0x16,r1
		shll8	r1
		mov.b	@(Screen_Flags,r6),r0
		tst	#SCREEN_FLAGS_DOUBLEPIXEL,r0
		mov	#-1,r2
		subc	r2,r1
		shll8	r1
		tst	#SCREEN_FLAGS_ENABLE,r0
		movt	r2
		shll2	r2
		shll	r2
		or	r2,r1
		mov	#BITMAPTYPE2,r0
		shll2	r0
		mov.l	r1,@(r0,r5)

		mov.b	@(Screen_DisplayType,r6),r0
		mov	#DISPLAY_SYSTEM_MASK,r2
		mov	r0,r1
		and	r2,r1
		mov.b	@(Screen_Flags,r6),r0
		mov.l	@(DisplayModeInfo_ImageControlILOR,r8),r1
		tst	#SCREEN_FLAGS_INTERLACE,r0
		movt	r2
		mov	#1,r0
		add	#-1,r2
		shll8	r0
		and	r2,r1
		or	r0,r1
		mov	#IMAGECONTROL,r0
		shll2	r0
		mov.l	r1,@(r0,r5)

		mov.l	@(DisplayModeInfo_FrameTotal,r8),r2
		mov.b	@(Screen_Flags,r6),r0
		tst	#SCREEN_FLAGS_INTERLACE,r0
		movt	r3
		add	#-1,r3
		shll16	r3
		and	r2,r3
		add	r3,r2
		mov	#FRAMETOTAL,r0
		shll2	r0
		mov.l	r2,@(r0,r5)

		mov.b	@(Screen_Flags,r6),r0
		mov.l	@(Screen_Height,r6),r1
		tst	#SCREEN_FLAGS_INTERLACE,r0
		movt	r2
		add	#-1,r2
		shld	r2,r1
		add	#-1,r1
		shll8	r1
		mov.l	scUpdateMode_HeightMask,r3
		shll2	r1
		and	r3,r1
		mov	#DIWSIZE*4,r0
		not	r3,r3
		mov.l	@(r0,r5),r2
		and	r3,r2
		or	r2,r1
		mov.l	r1,@(r0,r5)

		mov	#BORDERWINDOWX,r0
		mov.l	@(DisplayModeInfo_BorderWindowX,r8),r2
		shll2	r0
		mov.l	@(DisplayModeInfo_BorderWindowY,r8),r3
		mov.l	r2,@(r0,r5)
		mov	#BORDERWINDOWY,r0
		shll2	r0
		mov.l	r3,@(r0,r5)

		bsr	scUpdateBuffer
		nop
		bsr	scUpdateCenter
		nop

		lds.l	@r15+,pr
		rts
		mov.l	@r15+,r8


		.ALIGN	5
scUpdateMode_PVR2Regs:
		.long	PVR2REGS
scUpdateMode_HeightMask:
		.long	0x000ffc00
	
scUpdateMode_Table0:
		.byte	SCREEN_FLAGS_31KHZ							! 31kHz, yhires, xhires
		.byte	SCREEN_FLAGS_31KHZ|SCREEN_FLAGS_DOUBLEPIXEL				! 31kHz, yhires, xlores
		.byte	SCREEN_FLAGS_31KHZ|SCREEN_FLAGS_DOUBLESCAN				! 31kHz, ylores, xhires
		.byte	SCREEN_FLAGS_31KHZ|SCREEN_FLAGS_DOUBLESCAN|SCREEN_FLAGS_DOUBLEPIXEL	! 31kHz, ylores, xlores
		.byte	SCREEN_FLAGS_INTERLACE							! 15kHz, yhires, xhires
		.byte	SCREEN_FLAGS_INTERLACE|SCREEN_FLAGS_DOUBLEPIXEL				! 15kHz, yhires, xlores
		.byte	0									! 15kHz, ylores, xhires
		.byte	SCREEN_FLAGS_DOUBLEPIXEL						! 15kHz, ylores, xlores


DisplayModeInfo_ImageControlILOR	=	0
DisplayModeInfo_FrameTotal		=	4
DisplayModeInfo_BorderWindowX		=	8
DisplayModeInfo_BorderWindowY		=	12
DisplayModeInfo_DisplayOriginX		=	16
DisplayModeInfo_DisplayOriginY		=	20
DisplayModeInfo_HorizFreq		=	24
DisplayModeInfo_SIZEOF			=	28

scUpdateMode_Regs_15kHz_50Hz:			! 15kHz, 50Hz display
		.long	0x90				! ImageControl interlace-OR
		.long	(312<<16)+863			! FrameTotal
		.long	(141<<16)+843			! BorderWindowX
		.long	(22<<16)+308			! BorderWindowY
		.long	494				! DisplayCenterX
		.long	165				! DisplayCenterY
		.long	0				! HorizFreq

scUpdateMode_Regs_15kHz_60Hz:			! 15kHz, 60Hz display
		.long	0x50				! ImageControl interlace-OR
		.long	(262<<16)+857			! FrameTotal
		.long	(126<<16)+837			! BorderWindowX
		.long	(18<<16)+258			! BorderWindowY
		.long	484				! DisplayCenterX
		.long	138				! DisplayCenterY
		.long	0				! HorizFreq

scUpdateMode_Regs_31kHz_60Hz:			! 31kHz, 60Hz display
		.long	0				! (ImageControl interlace-OR)
		.long	(524<<16)+857			! FrameTotal
		.long	(126<<16)+837			! BorderWindowX
		.long	(40<<16)+520			! BorderWindowY
		.long	488				! DisplayCenterX
		.long	280				! DisplayCenterY
		.long	SCREEN_FLAGS_31KHZ		! HorizFreq



!------------------------------------------------------------------------
! void scUpdateBuffer(void)
!
! Updates the following registers:
!
! FRAMEBUF, FRAMEBUF2, DIWSIZE(modulo,rowlength)
!

		.ALIGN	5
scUpdateBuffer:
		mov.l	scUpdateBuffer_PVR2Regs,r5
		mova	_Screen,r0
		mov	r0,r6

		mov.l	@(Screen_Base,r6),r1
		mov	#FRAMEBUF,r0
		shll2	r0
		mov.l	r1,@(r0,r5)

		mov.b	@(Screen_Flags,r6),r0
		tst	#SCREEN_FLAGS_INTERLACE,r0
		bt	scUpdateBuffer_done

		mov.l	@(Screen_Pitch,r6),r2
		add	r2,r1
		mov	#FRAMEBUF2,r0
		shll2	r0
		mov.l	r1,@(r0,r5)
scUpdateBuffer_done:

		mov.l	@(Screen_Format,r6),r2
		mova	scUpdateBuffer_BPPShift,r0
		mov.l	@(Screen_Width,r6),r3
		mov.b	@(r0,r2),r2
		shlr2	r3
		mov.b	@(Screen_Flags,r6),r0
		shld	r2,r3
		mov.l	@(Screen_Pitch,r6),r2
		shlr2	r2
		tst	#SCREEN_FLAGS_INTERLACE,r0
		mov	r2,r1
		movt	r4
		sub	r3,r1
		add	#-1,r4
		and	r4,r2
		add	r2,r1
		add	#1,r1
		shll16	r1
		shll2	r1
		shll2	r1
		add	#-1,r3
		mov.l	scUpdateBuffer_Mask,r2
		or	r3,r1
		mov	#DIWSIZE*4,r0
		and	r2,r1
		not	r2,r2
		mov.l	@(r0,r5),r3
		and	r2,r3
		or	r3,r1
		rts
		mov.l	r1,@(r0,r5)	

				.ALIGN	4
scUpdateBuffer_PVR2Regs:	.long	PVR2REGS
scUpdateBuffer_Mask:		.long	0xfff003ff
scUpdateBuffer_BPPShift:	.byte	1, 1, 0, 2



!------------------------------------------------------------------------
! void scUpdateCenter(void)
!
! Updates the following registers:
!
! DIWSTARTH, DIWSTARTV
!

		.ALIGN	5
scUpdateCenter:
		mov.l	scUpdateCenter_PVR2Regs,r5
		mova	_Screen,r0
		mov	r0,r6

		mov.b	@(Screen_Flags,r6),r0
		mov.l	@(Screen_Width,r6),r1
		tst	#SCREEN_FLAGS_DOUBLEPIXEL,r0
		mov.l	@(Screen_XCenter,r6),r2
		subc	r3,r3
		shad	r3,r1
		mov.l	@(Screen_DisplayOriginX,r6),r3
		sub	r1,r2
		add	r3,r2
		mov	#DIWSTARTH,r0
		shll2	r0
		mov.l	r2,@(r0,r5)

		mov.b	@(Screen_Flags,r6),r0
		mov.l	@(Screen_Height,r6),r1
		tst	#SCREEN_FLAGS_31KHZ,r0
		movt	r2
		tst	#SCREEN_FLAGS_INTERLACE,r0
		movt	r3
		add	r3,r2
		tst	#SCREEN_FLAGS_DOUBLESCAN,r0
		subc	r3,r3
		add	r2,r3
		add	#-1,r3
		shld	r3,r1

		mov.l	@(Screen_YCenter,r6),r2
		sub	r1,r2
		mov.b	@(Screen_Flags,r6),r0
		tst	#SCREEN_FLAGS_31KHZ,r0
		subc	r3,r3
		shad	r3,r2
		mov.l	@(Screen_DisplayOriginY,r6),r3
		xor	#SCREEN_FLAGS_INTERLACE,r0
		add	r3,r2
		tst	#SCREEN_FLAGS_60HZ|SCREEN_FLAGS_INTERLACE,r0
		movt	r3
		add	r2,r3
		shll16	r3
		or	r3,r2
		mov	#DIWSTARTV,r0
		shll2	r0
		mov.l	r2,@(r0,r5)

		rts
		nop

				.ALIGN	2
scUpdateCenter_PVR2Regs:	.long	PVR2REGS



			.ALIGN	5

			.GLOBL	_Screen
_Screen:
			.byte	0,0,0,0
			.long	0x200000
			.long	640*4
			.long	640, 480
			.long	SCREEN_ARGB8888
			.long	1, 1
			.long	0
			.long	0, 0
			.long	0, 0
