! NOTE:	These serial driver routines are still crap.
! If PC tries to send a char when input FIFO is non-empty,
! the input FIFO will lock up (ie no more chars will be
! received/detected).
! Same thing with output FIFO, so sending code will
! wait for FIFO empty.
!
	
		.SECTION .text


!--------------------------------------------------------------
! void SerialWriteString(char* string)
!
! Writes the given ASCIIZ string to serial port
!
! in	r4	char*	string
!

		.GLOBL	_SerialWriteString
		.ALIGN	5
_SerialWriteString:
		mov.l	r8,@-r15
		sts.l	pr,@-r15

		mov	r4,r8
	
_SerialWriteString_nextChar:
		mov.b	@r8+,r4
		tst	r4,r4
		bt	_SerialWriteString_done

		bsr	_SerialWriteChar
		nop
	
		bra	_SerialWriteString_nextChar
		nop

_SerialWriteString_done:

		lds.l	@r15+,pr
		rts
		mov.l	@r15+,r8



!--------------------------------------------------------------
! void SerialWriteHexByte(uint8 value)
!
! Converts value to hexstring, and writes it to serial port
!
! in	r4	uint8	value
!

		.GLOBL	_SerialWriteHexByte
		.ALIGN	5
_SerialWriteHexByte:
		bra	_SerialWriteHexNibbles
		mov	#2,r5


	
!--------------------------------------------------------------
! void SerialWriteHexWord(uint16 value)
!
! Converts value to hexstring, and writes it to serial port
!
! in	r4	uint16	value
!

		.GLOBL	_SerialWriteHexWord
		.ALIGN	5
_SerialWriteHexWord:
		bra	_SerialWriteHexNibbles
		mov	#4,r5


	
!--------------------------------------------------------------
! void SerialWriteHexLong(uint32 value)
!
! Converts value to hexstring, and writes it to serial port
!
! in	r4	uint32	value
!

		.GLOBL	_SerialWriteHexLong
		.ALIGN	5
_SerialWriteHexLong:
		bra	_SerialWriteHexNibbles
		mov	#8,r5


	
!--------------------------------------------------------------
! void SerialWriteHexNibbles(uint32 value, int numHexNibbles)
!
! Converts value to hexstring, and writes numNibbles lowest nibbles of it to serial port
!
! in	r4	uint32	value
!	r5	int	numNibbles
!

		.GLOBL	_SerialWriteHexNibbles
		.ALIGN	5
_SerialWriteHexNibbles:
		mov.l	r9,@-r15
		mov.l	r8,@-r15
		sts.l	pr,@-r15

		mov	r4,r8
		mov	r5,r9	

		mov	#8,r0
		sub	r9,r0
		shll2	r0
		shld	r0,r8

_SerialWriteHexNibbles_nextDigit:
		rotl	r8
		rotl	r8
		rotl	r8
		mov	#10,r1
		rotl	r8
		mov	r8,r0
		and	#0xf,r0
		cmp/hs	r1,r0
		bf/s	_SerialWriteHexNibbles_decimalDigit
		add	#'0',r0

		add	#'A'-'0'-10,r0
_SerialWriteHexNibbles_decimalDigit:

		bsr	_SerialWriteChar
		mov	r0,r4

		dt	r9
		bf	_SerialWriteHexNibbles_nextDigit

		lds.l	@r15+,pr
		mov.l	@r15+,r8
		rts
		mov.l	@r15+,r9



!--------------------------------------------------------------
! void SerialWriteChar(char character)
!
! Writes the given character to serial port
!
! in	r4	char	character
!

		.GLOBL	_SerialWriteChar
		.ALIGN	5
_SerialWriteChar:
		mov	#0xe8,r1
		shll16	r1
		add	#0xc,r1
		mov.w	@(4,r1),r0
		tst	#0x20,r0
		bt	_SerialWriteChar

		mov.b	r4,@r1
		nop

		and	#0x9f,r0
		mov.w	r0,@(4,r1)

_SerialWriteChar_waitForTransmission:
		mov.w	@(4,r1),r0
		tst	#0x40,r0
		bt	_SerialWriteChar_waitForTransmission
		rts
		nop



!--------------------------------------------------------------
! uint8 SerialReadUChar(void)
!
! Reads a character from serial port
!
! out	r0	int8	character
!

		.GLOBL	_SerialReadUChar
		.ALIGN	5
_SerialReadUChar:
		mov	#0xe8,r1
		shll16	r1
		add	#0xc,r1
		mov.w	@(0x10,r1),r0
		tst	#0x1f,r0
		bt	_SerialReadUChar

		mov.b	@(8,r1),r0
		extu.b	r0,r2

		mov.w	@(4,r1),r0
		and	#0x6d,r0
		mov.w	r0,@(4,r1)

		rts
		mov	r2,r0



!--------------------------------------------------------------
! bool SerialCheckInput(void)
!
! Check if input FIFO is non-empty
!
! out	r0	bool	true if there is a character available
!

		.GLOBL	_SerialCheckInput
		.ALIGN	5
_SerialCheckInput:
		mov	#0xe8,r1
		shll16	r1
		add	#0xc,r1
		mov.w	@(0x10,r1),r0
		tst	#0x1f,r0
		subc	r0,r0
		rts
		add	#1,r0
