
Math_PI		=	0x40490fdb		! 3.141592
Math_FSCASCALE	=	0x4622f983		! 65536 / 2*PI = 10430.378

Mat4x4_SIZEOF	=	16*4

		.SECTION .text


!-------------------------------------------------------
! void XMat4x4_load(Mat4x4* matrix)
!
! Loads XMTRX from memory
!
! In	r4	Mat4x4	New matrix
! Out	XMTRX	Mat4x4	Loaded matrix
!

		.GLOBL	_XMat4x4_load
		.ALIGN	5
_XMat4x4_load:
		fschg
		fmov.d	@r4,xd0
		add	#32,r4
		pref	@r4
		add	#-(32-8),r4
		fmov.d	@r4+,xd2
		fmov.d	@r4+,xd4
		fmov.d	@r4+,xd6
		fmov.d	@r4+,xd8
		fmov.d	@r4+,xd10
		fmov.d	@r4+,xd12
		fmov.d	@r4+,xd14
		rts
		fschg



!-------------------------------------------------------
! void XMat4x4_store(Mat4x4* matrix)
!
! Stores XMTRX to memory
!
! In	r4	Mat4x4	Address to store matrix at
!	XMTRX	Mat4x4	Current matrix
!

		.GLOBL	_XMat4x4_store
		.ALIGN	5
_XMat4x4_store:
		fschg
		add	#64-8,r4
		fmov.d	xd14,@r4
		add	#-32,r4
		pref	@r4
		add	#32,r4
		fmov.d	xd12,@-r4
		fmov.d	xd10,@-r4
		fmov.d	xd8,@-r4
		fmov.d	xd6,@-r4
		fmov.d	xd4,@-r4
		fmov.d	xd2,@-r4
		fmov.d	xd0,@-r4
		rts
		fschg



!-------------------------------------------------------
! void XMat4x4_setIdentity(void)
!
! Sets XMTRX to identity
!
! Out	XMTRX	Mat4x4	Identity matrix
!

		.GLOBL	_XMat4x4_setIdentity
		.ALIGN	5
_XMat4x4_setIdentity:
		frchg
		fldi1	fr0
		fschg
		fldi0	fr1
		fldi0	fr2
		fldi0	fr3
		fldi0	fr4
		fldi1	fr5
		fmov	dr2,dr6
		fmov	dr2,dr8
		fmov	dr0,dr10
		fmov	dr2,dr12
		fmov	dr4,dr14
		fschg
		rts
		frchg



!-----------------------------------------------------
! void XMat4x4_setRotation(Vec4* rotAngles)
!
! Set XMTRX to 3x3 rotation matrix - Y,X,Z rotation order
!
! In	r4	Vec4	Rotation angles (only X,Y,Z used)
! Out	XMTRX	Mat4x4	Rotation matrix
!

		.GLOBL	_XMat4x4_setRotation
		.ALIGN	5
_XMat4x4_setRotation:
		mov.l	r8,@-r15
		sts.l	pr,@-r15
		mov	r4,r8

		add	#8,r4
		bsr	_XMat4x4_setZRotation
		fmov.s	@r4,fr4

		mova	_XMat4x4_setRotation_MatBufZ,r0
		bsr	_XMat4x4_store
		mov	r0,r4

		bsr	_XMat4x4_setXRotation
		fmov.s	@r8+,fr4

		mova	_XMat4x4_setRotation_MatBufX,r0
		bsr	_XMat4x4_store
		mov	r0,r4

		bsr	_XMat4x4_setYRotation
		fmov.s	@r8,fr4

		mova	_XMat4x4_setRotation_MatBufX,r0
		bsr	_XMat4x4_applyMatrix
		mov	r0,r4

		mova	_XMat4x4_setRotation_MatBufZ,r0
		bsr	_XMat4x4_applyMatrix
		mov	r0,r4

		lds.l	@r15+,pr
		mov.l	@r15+,r8
		rts
		nop

				.ALIGN	5
_XMat4x4_setRotation_MatBufX:	.SPACE	Mat4x4_SIZEOF
_XMat4x4_setRotation_MatBufZ:	.SPACE	Mat4x4_SIZEOF

	

!-----------------------------------------------------
! void XMat4x4_setXRotation(float angle)
!
! Sets XMTRX to X rotation
!
! In	fr4	float	X rotation angle
! Out	XMTRX	Mat4x4	Rotation matrix
!

		.GLOBL	_XMat4x4_setXRotation
		.ALIGN	5
_XMat4x4_setXRotation:
		mova	_XMat4x4_setXRotation_AngleScale,r0
		fschg
		fmov.d	@r0,xd0
		fmov	dr4,xd4
		frchg
		fldi0	fr2
		fschg
		fldi0	fr3
		fmul	fr0,fr4
		fldi0	fr7
		fldi0	fr8
		ftrc	fr4,fpul
		fldi0	fr12
		fldi0	fr13
		fsca	fpul,dr0
		fldi0	fr4
		fldi0	fr11
		fldi0	fr14
		fldi1	fr15
		fmov	fr1,fr5
		fmov	fr1,fr10
		fmov	fr0,fr9
		fneg	fr9
		fmov	fr0,fr6
		fldi1	fr0
		fldi0	fr1
		rts
		frchg

		.ALIGN	3

_XMat4x4_setXRotation_AngleScale: .long Math_FSCASCALE,0



!-----------------------------------------------------
! void XMat4x4_setYRotation(float angle)
!
! Sets XMTRX to Y rotation
!
! In	fr4	float	Y rotation angle
! Out	XMTRX	Mat4x4	Rotation matrix
!

		.GLOBL	_XMat4x4_setYRotation
		.ALIGN	5
_XMat4x4_setYRotation:
		mova	_XMat4x4_setYRotation_AngleScale,r0
		fschg
		fmov.d	@r0,xd0
		fmov	dr4,xd4
		frchg
		fldi0	fr3
		fschg
		fldi1	fr5
		fmul	fr0,fr4
		fldi0	fr6
		fldi0	fr7
		ftrc	fr4,fpul
		fldi0	fr12
		fldi0	fr13
		fsca	fpul,dr0
		fldi0	fr4
		fldi0	fr9
		fldi0	fr11
		fldi0	fr14
		fldi1	fr15
		fmov	fr1,fr10
		fmov	fr0,fr8
		fneg	fr8
		fmov	fr0,fr2
		fmov	fr1,fr0

		fldi0	fr1
		rts
		frchg

		.ALIGN	3

_XMat4x4_setYRotation_AngleScale: .long Math_FSCASCALE,0



!-----------------------------------------------------
! void XMat4x4_setZRotation(float angle)
!
! Set XMTRX to Z rotation
!
! In	fr4	float	Z rotation angle
! Out	XMTRX	Mat4x4	Rotation matrix
!

		.GLOBL	_XMat4x4_setZRotation
		.ALIGN	5
_XMat4x4_setZRotation:
		mova	_XMat4x4_setZRotation_AngleScale,r0
		fschg
		fmov.d	@r0,xd0
		fmov	dr4,xd4
		frchg
		fldi0	fr2
		fschg
		fldi0	fr3
		fmul	fr0,fr4
		fldi1	fr10
		fldi0	fr11
		ftrc	fr4,fpul
		fsca	fpul,dr4
		fschg
		fmov	dr2,dr6
		fmov	dr2,dr8
		fmov	dr2,dr12
		fldi0	fr14
		fldi1	fr15
		fschg
		fmov	fr5,fr0
		fmov	fr4,fr1
		fneg	fr4
		rts
		frchg

		.ALIGN	3

_XMat4x4_setZRotation_AngleScale: .long Math_FSCASCALE,0

	

!-----------------------------------------------------
! void XMat4x4_setScale(Vec4* scales)
!
! Set XMTRX to non-uniform scale matrix
!
! In	r4	Vec4	Scales
! Out	XMTRX	Mat4x4	Scale matrix
!

		.GLOBL	_XMat4x4_setScale
		.ALIGN	5
_XMat4x4_setScale:
		pref	@r4
		frchg
		fldi0	fr1
		fschg
		fldi0	fr2
		fldi0	fr3
		fldi0	fr4
		fmov	dr2,dr6
		fmov	dr2,dr8
		fldi0	fr11
		fmov	dr2,dr12
		fldi0	fr14
		fschg
		fmov.s	@r4+,fr0
		fmov.s	@r4+,fr5
		fmov.s	@r4+,fr10
		fmov.s	@r4+,fr15
		rts
		frchg

	
	
!-----------------------------------------------------
! void XMat4x4_setTranslation(Vec4* translation)
!
! Set XMTRX to translation
!
! In	r4	Vec4	translation
! Out	XMTRX	Mat4x4	Translation matrix
!

		.GLOBL	_XMat4x4_setTranslation
		.ALIGN	5
_XMat4x4_setTranslation:
		pref	@r4
		frchg
		fldi1	fr0
		fschg
		fldi0	fr1
		fldi0	fr2
		fldi0	fr3
		fldi0	fr4
		fldi1	fr5
		fmov	dr2,dr6
		fmov	dr2,dr8
		fmov	dr0,dr10
		fmov.d	@r4+,dr12
		fmov.d	@r4+,dr14
		fschg
		rts
		frchg

	

!-----------------------------------------------------
! void XMat4x4_applyTranslation(Vec4* translation)
!
! Apply translation to XMTRX
!
! In	r4	Vec4	Translation vector
!	XMTRX	Mat4x4	Current matrix
! Out	XMTRX	Mat4x4	Combined matrix
!

		.GLOBL	_XMat4x4_applyTranslation
		.ALIGN	5
_XMat4x4_applyTranslation:
		pref	@r4
		fschg
		fmov	xd12,dr4
		fmov	xd14,dr6
		fmov.d	@r4+,dr0
		fmov.d	@r4+,dr2
		fschg
		fadd	fr0,fr4
		fadd	fr1,fr5
		fadd	fr2,fr6
		fadd	fr3,fr7
		fschg
		fmov	dr4,xd12
		fmov	dr6,xd14
		rts
		fschg



!-----------------------------------------------------
! void XMat4x4_applyScale(Vec4* scales)
!
! Apply nonuniform scales to XMTRX
!
! In	r4	Vec4	Scale factors
!	XMTRX	Mat4x4	Current matrix
! Out	XMTRX	Mat4x4	Combined matrix
!

		.GLOBL	_XMat4x4_applyScale
		.ALIGN	5
_XMat4x4_applyScale:
		fschg
		fmov	xd0,dr4
		fmov.d	@r4+,xd0
		fmov	xd2,dr6
		frchg
		fmov.d	@r4+,dr2
		fschg
		
		fmul	fr0,fr4
		fmul	fr0,fr8
		fmul	fr0,fr12
		fmul	fr1,fr5
		fmul	fr1,fr9
		fmul	fr1,fr13
		fmul	fr2,fr6
		fmul	fr2,fr10
		fmul	fr2,fr14

		fschg
		fmov	dr4,xd0
		fmul	fr3,fr7
		fmov	dr6,xd2
		fmul	fr3,fr11
		fmov	xd4,dr4
		fmul	fr3,fr15
		fmov	xd6,dr6
		fschg
		fmul	fr4,fr0
		fmul	fr5,fr1
		fmul	fr6,fr2
		fmul	fr7,fr3
		fschg

		fmov	xd0,dr4
		fmov	xd2,dr6
		fschg
		rts
		frchg



!-----------------------------------------------------
! void XMat4x4_apply2DProjectionShearShuffle(Vec2* xyScale, Vec2* xyCenter)
!
! Apply 2d scale + shear + XYZW -> ZXYW rotation to XMTRX
!
! In	r4	Vec2	Scale factors
!	r5	Vec2	Projection center
!	XMTRX	Mat4x4	Current matrix
! Out	XMTRX	Mat4x4	Combined matrix
!

		.GLOBL	_XMat4x4_apply2DProjectionShearShuffle
		.ALIGN	5
_XMat4x4_apply2DProjectionShearShuffle:
		fschg
		fmov	xd2,dr0
		fmov.d	@r4+,xd2
		fmov	xd6,dr2
		frchg
		fneg	fr3
		fschg
		fmul	fr3,fr1
		fmul	fr3,fr5
		fmul	fr3,fr9
		fmul	fr3,fr13
		fmul	fr2,fr0
		fmul	fr2,fr4
		fmul	fr2,fr8
		fmul	fr2,fr12

		fmov.s	@r5+,fr11
		fmov.s	@r5+,fr15

		fschg
		fmov	xd0,dr2
		fmov	xd2,dr6
		fschg

		fmov	fr2,fr3
		fmul	fr11,fr3
		fmov	fr6,fr7
		fmul	fr11,fr7
		fadd	fr3,fr0
		fadd	fr7,fr4
		fmov	fr10,fr3
		fmul	fr11,fr3
		fmov	fr14,fr7
		fmul	fr11,fr7
		fadd	fr3,fr8
		fadd	fr7,fr12

		fmov	fr2,fr3
		fmul	fr15,fr3
		fmov	fr6,fr7
		fmul	fr15,fr7
		fadd	fr3,fr1
		fadd	fr7,fr5
		fmov	fr10,fr3
		fmul	fr15,fr3
		fmov	fr14,fr7
		fmul	fr15,fr7
		fadd	fr3,fr9
		fadd	fr7,fr13

		fmov	fr2,fr3
		fmov	fr6,fr7
		fmov	fr10,fr11
		fmov	fr14,fr15
		fmov	fr1,fr2
		fmov	fr5,fr6
		fmov	fr9,fr10
		fmov	fr13,fr14
		fmov	fr0,fr1
		fmov	fr4,fr5
		fmov	fr8,fr9
		fmov	fr12,fr13
		fmov	fr3,fr0
		fmov	fr7,fr4
		fmov	fr11,fr8
		fmov	fr15,fr12

		fldi0	fr3
		fldi0	fr7
		fldi0	fr11
		fldi1	fr15
		rts
		frchg



!-----------------------------------------------------
! void XMat4x4_applyMatrix(Mat4x4* matrix)
!
! Apply 4x4 matrix to XMTRX
!
! In	r4	Mat4x4	Second matrix
!	XMTRX	Mat4x4	Current matrix
! Out	XMTRX	Mat4x4	Combined matrix
!

		.GLOBL	_XMat4x4_applyMatrix
		.ALIGN	5
_XMat4x4_applyMatrix:
		mov	r15,r0
		pref	@r4
		and	#0xf0,r0
		fschg
		mov	r15,r7
		mov	r0,r15
		fmov.d	dr14,@-r15
		fmov.d	dr12,@-r15

		fmov.d	@r4,dr0
		add	#32,r4
		pref	@r4
		add	#-(32-8),r4
		fmov.d	@r4+,dr2
		fmov.d	@r4+,dr4
		fmov.d	@r4+,dr6
		fmov.d	@r4+,dr8
		fmov.d	@r4+,dr10
		fmov.d	@r4+,dr12
		fmov.d	@r4+,dr14
		frchg

		ftrv	xmtrx,fv0
		ftrv	xmtrx,fv4
		ftrv	xmtrx,fv8
		ftrv	xmtrx,fv12

		frchg
		fmov.d	@r15+,dr12
		fmov.d	@r15,dr14
		mov	r7,r15
		rts
		fschg



!-----------------------------------------------------
! void XMat4x4_applyLightMatrix(Mat4x4* matrix)
!
! Apply 4x4 light matrix to XMTRX
!
! In	r4	Mat4x4	Second matrix
!	XMTRX	Mat4x4	Current matrix
! Out	XMTRX	Mat4x4	Combined matrix
!

		.GLOBL	_XMat4x4_applyLightMatrix
		.ALIGN	5
_XMat4x4_applyLightMatrix:
		mov	r15,r0
		pref	@r4
		and	#0xf0,r0
		fschg
		mov	r15,r7
		mov	r0,r15
		fmov.d	dr14,@-r15
		fmov.d	dr12,@-r15

		fmov.d	@r4,dr0
		add	#32,r4
		pref	@r4
		add	#-(32-8),r4
		fmov.d	@r4+,dr2
		fmov.d	@r4+,dr4
		fmov.d	@r4+,dr6
		fmov.d	@r4+,dr8
		fmov.d	@r4+,dr10
		fmov.d	@r4+,dr12
		fmov.d	@r4+,dr14

		fldi0	fr12
		fldi0	fr13
		fldi0	fr14
		fldi0	fr15
		frchg
		fldi0	fr12
		fldi0	fr13
		fldi0	fr14
		fldi0	fr15
	
		ftrv	xmtrx,fv0
		ftrv	xmtrx,fv4
		ftrv	xmtrx,fv8
		ftrv	xmtrx,fv12

		frchg
		fmov.d	@r15+,dr12
		fmov.d	@r15,dr14
		mov	r7,r15
		rts
		fschg



!-----------------------------------------------------
! void XMat4x4_applyLightMatrixConversion(void)
!
! Apply scale/shift to bring [0.0, 1.0] range to [-2^31, +2^31]
!
! in	XMTRX	Mat4x4	Current matrix
! out	XMTRX	Mat4x4	Light matrix
!

		.GLOBL	_XMat4x4_applyLightMatrixConversion
		.ALIGN	5
_XMat4x4_applyLightMatrixConversion:

		mova	_XMat4x4_applyLightMatrixConversion_Factors,r0
		fschg
		fmov.d	@r0,dr4
		fmov	xd0,dr0
		fmov	xd2,dr2
		fschg

		fmul	fr4,fr0
		fmul	fr4,fr1
		fmul	fr4,fr2
		fmul	fr4,fr3

		frchg
		fmov.s	@r0+,fr0
		fmov.s	@r0+,fr1
		fmul	fr0,fr4
		fmul	fr0,fr5
		fmul	fr0,fr6
		fmul	fr0,fr7
		fmul	fr0,fr8
		fmul	fr0,fr9
		fmul	fr0,fr10
		fmul	fr0,fr11
		fmul	fr0,fr12
		fmul	fr0,fr13
		fmul	fr0,fr14
		fmul	fr0,fr15
		fadd	fr1,fr12
		fadd	fr1,fr13
		fadd	fr1,fr14
		fadd	fr1,fr15

		fschg
		fmov	xd0,dr0
		fmov	xd2,dr2
		fschg
		rts
		frchg

		.ALIGN	3
_XMat4x4_applyLightMatrixConversion_Factors:
		.LONG	0x4f000000
		.LONG	0 !0xcf000000
