! Test example for the Render module
!
! Draws a morphing donut with 2*DONUTSTEPS1*DONUTSTEPS2 stripped triangles
!
! Border colours:
! VSync interrupt sets red border (at top of screen)
! RenderDone interrupt sets yellow border (position varies)
! Vertex registration done sets green border (steadily somewhere on screen)
!
! Currently, the code manages to push out approx. 55k striptris at 60fps.
! It is possible to do the CPU/TA work at least twice as fast.
! The rendering hardware will not be able to draw more than 60k-80k
! striptris per frame though (exact figure depends on size and orientation
! of the tris)
!

DONUTSTEPS1	=	130
DONUTSTEPS2	=	200

Donut0Vertices		= 0x8c000000 + 8*1024*1024		! Base address of donut 0 vertex data (temp)
Donut1Vertices		= 0x8c000000 + 9*1024*1024		! Base address of donut 1 vertex data (temp)
Donut0VertexNormals	= 0x8c000000 + 10*1024*1024		! Base address of donut 0 vertexnormals (temp)
Donut1VertexNormals	= 0x8c000000 + 11*1024*1024		! Base address of donut 1 vertexnormals (temp)
DonutSuperStrip		= 0x8c000000 + 4*1024*1024		! Base address of donut superstrip data
DonutMorphVertices	= 0x8c000000 + 12*1024*1024		! Base address of donut morph-vertex data
DonutVertices		= 0x8c000000 + 8*1024*1024		! Output result from morphing

		.INCLUDE "PVR2_Regs.i"
		.INCLUDE "Render.i"
		.INCLUDE "SuperStrip.i"

	
		.SECTION .text

		.GLOBL	_Draw3D
		.ALIGN	5
_Draw3D:
		sts.l	pr,@-r15

		bsr	ClearGfxMem
		nop

! First, create the source data

		mov.l	_Draw3D_Donut0Vertices,r4
		bsr	GenerateDonut0Vertices
		nop

		mov.l	_Draw3D_Donut1Vertices,r4
		bsr	GenerateDonut1Vertices
		nop

		mov.l	_Draw3D_DonutSuperStrip,r4
		bsr	GenerateDonutSuperStrip
		nop

		mov.l	_Draw3D_DonutSuperStrip,r4
		mov.l	_Draw3D_Donut0Vertices,r5
		mov.l	_Draw3D_Donut0VertexNormals,r6
		mov.l	_Draw3D_UberStrip_calcVertexNormals,r0
		mov.l	_Draw3D_NumDonutVertices,r7
		jsr	@r0
		nop
	
		mov.l	_Draw3D_DonutSuperStrip,r4
		mov.l	_Draw3D_Donut1Vertices,r5
		mov.l	_Draw3D_Donut1VertexNormals,r6
		mov.l	_Draw3D_UberStrip_calcVertexNormals,r0
		mov.l	_Draw3D_NumDonutVertices,r7
		jsr	@r0
		nop
	
		mov.l	_Draw3D_Donut0Vertices,r4
		mov.l	_Draw3D_Donut1Vertices,r5
		mov.l	_Draw3D_Donut0VertexNormals,r6
		mov.l	_Draw3D_NumDonutVertices,r2
		mov.l	_Draw3D_Donut1VertexNormals,r7
		mov.l	_Draw3D_DonutMorphVertices,r1
		mov.l	_Draw3D_CalculateDonutMorphVertices,r0
		mov.l	r2,@-r15
		jsr	@r0
		mov.l	r1,@-r15
		add	#8,r15

! Now initialize renderer
		
		mov.l	_Draw3D_rdInit,r0
		jsr	@r0
		nop

		mov.l	_Draw3D_rdSetVSyncUserCallback,r1
		mova	SetBorderRed,r0
		jsr	@r1
		mov	r0,r4
	
		mov.l	_Draw3D_rdSetRenderDoneUserCallback,r1
		mova	SetBorderYellow,r0
		jsr	@r1
		mov	r0,r4
	
		mov.l	_Draw3D_SetIMASK,r0
		jsr	@r0
		mov	#0,r4

		add	#-8,r15
		mov.l	_Draw3D_rdSetupNativeBuffers,r0
		mov	#RENDER_BUCKETSIZE_16,r4
		mov	#RENDER_BUCKETSIZE_0,r5
		mov	#RENDER_BUCKETSIZE_16,r6
		mov	#RENDER_BUCKETSIZE_0,r7
		mov.l	_Draw3D_AllocationBufferSize,r1
		mov.l	r7,@(0,r15)
		jsr	@r0
		mov.l	r1,@(4,r15)
		add	#8,r15

		mov.l	_Draw3D_rdSetupVertexBuffer,r0		! Not really needed in this case
		mov.l	_Draw3D_OpaqueBufferStart,r5
		mov.l	_Draw3D_OpaqueBufferLength,r6
		jsr	@r0
		mov	#RENDER_LIST_OPAQUE,r4

		mov.l	_Draw3D_rdSetupVertexBuffer,r0		! And neither is this
		mov.l	_Draw3D_TranslucentBufferStart,r5
		mov.l	_Draw3D_TranslucentBufferLength,r6
		jsr	@r0
		mov	#RENDER_LIST_TRANSLUCENT,r4

! Render stuff

Draw3D_frame:
		bsr	DrawFrame
		nop

		mov.l	_Draw3D_SerialCheckInput,r0
		jsr	@r0
		nop

		tst	r0,r0
		bt	Draw3D_frame
	
		mov.l	_Draw3D_rdShutdown,r0
		jsr	@r0
		nop

		lds.l	@r15+,pr
		rts
		nop

				.ALIGN	5
_Draw3D_rdInit:			.long	_rdInit
_Draw3D_rdSetupNativeBuffers:	.long	_rdSetupNativeBuffers
_Draw3D_AllocationBufferSize:	.long	2*1024*1024+768*1024
_Draw3D_rdShutdown:		.long	_rdShutdown
_Draw3D_rdSetupVertexBuffer:	.long	_rdSetupVertexBuffer
_Draw3D_SetIMASK:		.long	_SetIMASK
_Draw3D_OpaqueBufferStart:	.long	0x8c040000
_Draw3D_OpaqueBufferLength:	.long	0x00010000
_Draw3D_TranslucentBufferStart:	.long	0x8c050000
_Draw3D_TranslucentBufferLength: .long	0x00010000
_Draw3D_SerialCheckInput:	.long	_SerialCheckInput
_Draw3D_rdSetVSyncUserCallback:	.long	_rdSetVSyncUserCallback
_Draw3D_rdSetRenderDoneUserCallback: .long _rdSetRenderDoneUserCallback
_Draw3D_Donut0Vertices:		.long	Donut0Vertices
_Draw3D_Donut1Vertices:		.long	Donut1Vertices
_Draw3D_DonutSuperStrip:	.long	DonutSuperStrip
_Draw3D_Donut0VertexNormals:	.long	Donut0VertexNormals
_Draw3D_Donut1VertexNormals:	.long	Donut1VertexNormals
_Draw3D_DonutMorphVertices:	.long	DonutMorphVertices
_Draw3D_NumDonutVertices:	.long	DONUTSTEPS1*DONUTSTEPS2
_Draw3D_UberStrip_calcVertexNormals:	.long _UberStrip_calcVertexNormals
_Draw3D_CalculateDonutMorphVertices:	.long	CalculateDonutMorphVertices

		.ALIGN	5
SetBorderRed:						! Set the border colour to RED
		mov.l	SetBorderRed_BorderColor,r0
		mov.l	SetBorderRed_Colour,r1
		rts
		mov.l	r1,@r0

				.ALIGN	3	
SetBorderRed_BorderColor:	.LONG	PVR2REGS+BORDERCOLOR*4
SetBorderRed_Colour:		.LONG	0x00ff0000

SetBorderGreen:						! Set the border colour to GREEN
		mov.l	SetBorderGreen_BorderColor,r0
		mov.l	SetBorderGreen_Colour,r1
		rts
		mov.l	r1,@r0

				.ALIGN	3	
SetBorderGreen_BorderColor:	.LONG	PVR2REGS+BORDERCOLOR*4
SetBorderGreen_Colour:		.LONG	0x0000ff00
		
		.ALIGN	5
SetBorderYellow:					! Set the border colour to YELLOW
		mov.l	SetBorderYellow_BorderColor,r0
		mov.l	SetBorderYellow_Colour,r1
		rts
		mov.l	r1,@r0

				.ALIGN	3	
SetBorderYellow_BorderColor:	.LONG	PVR2REGS+BORDERCOLOR*4
SetBorderYellow_Colour:		.LONG	0x00ffff00

			
ClearGfxMem:						! Zero out all graphics memory
		mov	#0x80/4,r1
		shll16	r1
		mov	#0xa5,r0
		shll16	r0
		shll8	r0
		mov	#0,r2
ClearGfxMem_long:
		dt	r1
		mov.l	r2,@r0
		bf/s	ClearGfxMem_long
		add	#4,r0
		rts
		nop



!----------------------------------------------------------------------
! void GenerateDonut0Vertices(void* output)
!
! Generates the vertices of a normal donut
!
! in	r4	void*	Destination for vertices
!

		.ALIGN	5
GenerateDonut0Vertices:
		mov.l	r11,@-r15
		mov.l	r10,@-r15
		mov.l	r9,@-r15
		mov.l	r8,@-r15

		mov.l	GenerateDonut_DonutSteps1,r10
		mov.l	GenerateDonut_DonutSteps2,r11

		mov	r4,r14
	
		lds	r10,fpul
		mova	GenerateDonut_DonutRadius1,r0
		float	fpul,fr0
		fmov.s	@r0,fr2
		lds	r11,fpul
		mova	GenerateDonut_DonutRadius2,r0
		float	fpul,fr1
		fmov.s	@r0,fr3
		mova	GenerateDonut_65536,r0
		fmov.s	@r0,fr4
		fmov	fr4,fr5
		fdiv	fr0,fr4
		fdiv	fr1,fr5

		mov	#0,r7
GenerateDonut0Vertices_vtx2:
		mov	#0,r6
GenerateDonut0Vertices_vtx1:
		lds	r6,fpul
		float	fpul,fr0
		fmul	fr4,fr0
		ftrc	fr0,fpul
		fsca	fpul,dr0

		fmov	fr0,fr9			! (x,y,z) = (sin(a) * radius1, cos(a) * radius1, <0>)
		fmov	fr1,fr8
		fmul	fr2,fr8
		fmul	fr2,fr9
	
		fadd	fr3,fr8			! (x',y',z') = (x + radius2, y, <z>)

		lds	r7,fpul
		float	fpul,fr1
		fmul	fr5,fr1
		ftrc	fr1,fpul
		fsca	fpul,dr0

		fmov	fr8,fr10		! (x'',y'',z'') = (cos(b) * x', y', sin(b) * x')
		fmul	fr1,fr8
		fmul	fr0,fr10

		fneg	fr10

		mov.l	GenerateDonut0Vertices_ColourFill,r0
		mov	#0,r1
		add	#16,r4
		mov.l	r0,@-r4
		fmov.s	fr10,@-r4
		fmov.s	fr9,@-r4
		fmov.s	fr8,@-r4
		add	#Vtx_SIZEOF,r4

		add	#1,r6
		cmp/hs	r10,r6
		bf	GenerateDonut0Vertices_vtx1
		
		add	#1,r7
		cmp/hs	r11,r7
		bf	GenerateDonut0Vertices_vtx2

		mov.l	@r15+,r8
		mov.l	@r15+,r9
		mov.l	@r15+,r10
		rts
		mov.l	@r15+,r11

		.ALIGN	2
GenerateDonut0Vertices_ColourFill:	.LONG	0x3f000000
	

!----------------------------------------------------------------------
! void GenerateDonut1Vertices(void* output)
!
! Generates the vertices of a deformed donut
!
! in	r4	void*	Destination for vertices
!

		.ALIGN	5
GenerateDonut1Vertices:
		fmov.s	fr15,@-r15
		fmov.s	fr14,@-r15
		fmov.s	fr13,@-r15
		fmov.s	fr12,@-r15
		mov.l	r11,@-r15
		mov.l	r10,@-r15
		mov.l	r9,@-r15
		mov.l	r8,@-r15

		mov.l	GenerateDonut_DonutSteps1,r10
		mov.l	GenerateDonut_DonutSteps2,r11

		mov	r4,r14
	
		lds	r10,fpul
		mova	GenerateDonut_DonutRadius1,r0
		float	fpul,fr0
		fmov.s	@r0,fr2
		lds	r11,fpul
		mova	GenerateDonut_DonutRadius2,r0
		float	fpul,fr1
		fmov.s	@r0,fr3
		mova	GenerateDonut_65536,r0
		fmov.s	@r0,fr4
		fmov	fr4,fr5
		fdiv	fr0,fr4
		fdiv	fr1,fr5

		mov	#0,r7
GenerateDonut1Vertices_vtx2:
		mov	#0,r6
GenerateDonut1Vertices_vtx1:
		lds	r6,fpul
		float	fpul,fr0
		fmul	fr4,fr0
		ftrc	fr0,fpul
		fsca	fpul,dr0

		fmov	fr0,fr9			! (x,y,z) = (sin(a) * radius1, cos(a) * radius1, <0>)
		fmov	fr1,fr8
		fmul	fr2,fr8
		fmul	fr2,fr9

		lds	r6,fpul			! scale = 1.f + 0.25f * (cos(a * 15.f) + 1.f) * (cos(b * 15.f) + 1.f)
		mova	GenerateDonut1Vertices_Factors,r0
		fmov.s	@r0+,fr12
		float	fpul,fr13
		fmul	fr4,fr13
		fmul	fr12,fr13
		ftrc	fr13,fpul
		fsca	fpul,dr14
		fldi1	fr12
		fadd	fr12,fr15
		lds	r7,fpul
		fmov.s	@r0+,fr12
		float	fpul,fr13
		fmul	fr5,fr13
		fmul	fr12,fr13
		ftrc	fr13,fpul
		fmov	fr15,fr13
		fsca	fpul,dr14
		fldi1	fr12
		fadd	fr12,fr15
		fmul	fr13,fr15
		fmov.s	@r0+,fr12
		fmul	fr12,fr15
		fldi1	fr12
		fadd	fr12,fr15
		fmul	fr15,fr8		! x *= scale
		fmul	fr15,fr9		! y *= scale
	
		fadd	fr3,fr8			! (x',y',z') = (x + radius2, y, <z>)

		lds	r7,fpul
		float	fpul,fr1
		fmul	fr5,fr1
		ftrc	fr1,fpul
		fsca	fpul,dr0

		fmov	fr8,fr10		! (x'',y'',z'') = (cos(b) * x', y', sin(b) * x')
		fmul	fr1,fr8
		fmul	fr0,fr10

		fneg	fr10

		mov.l	GenerateDonut1Vertices_ColourFill,r0
		mov	#0,r1
		add	#16,r4
		mov.l	r0,@-r4
		fmov.s	fr10,@-r4
		fmov.s	fr9,@-r4
		fmov.s	fr8,@-r4
		add	#Vtx_SIZEOF,r4

		add	#1,r6
		cmp/hs	r10,r6
		bf	GenerateDonut1Vertices_vtx1
		
		add	#1,r7
		cmp/hs	r11,r7
		bf	GenerateDonut1Vertices_vtx2

		mov.l	@r15+,r8
		mov.l	@r15+,r9
		mov.l	@r15+,r10
		mov.l	@r15+,r11
		fmov.s	@r15+,fr12
		fmov.s	@r15+,fr13
		fmov.s	@r15+,fr14
		rts
		fmov.s	@r15+,fr15

					.ALIGN	4
GenerateDonut1Vertices_ColourFill:	.LONG	0x3f000000
GenerateDonut1Vertices_Factors:		.SINGLE 15.0, 15.0, 0.25
	

!----------------------------------------------------------------------
! void GenerateDonutSuperStrip(void* output)
!
! Generates a superstrip for an entire donut
!
! in	r4	void*	Destination for superstrip
!

		.ALIGN	5
GenerateDonutSuperStrip:
		mov.l	r14,@-r15
		mov.l	r13,@-r15
		mov.l	r12,@-r15
		mov.l	r11,@-r15
		mov.l	r10,@-r15
		mov.l	r9,@-r15
		mov.l	r8,@-r15

		mov.l	GenerateDonut_DonutSteps1,r10
		mov.l	GenerateDonut_DonutSteps2,r11

		mov	#0xe0,r12
		mov	#0xf0,r13
		shll16	r12
		shll16	r13
		shll8	r12
		shll8	r13

		mov	#16,r8
		mul.l	r10,r8
		sts	macl,r9

		mov.l	r11,@r4
		add	#4,r4

		mov	#0,r7
GenerateDonutSuperStrip_strip:

		mov	r10,r0
		add	#1,r0
		add	r0,r0
		mov.l	r0,@r4
		add	#4,r4

		mov	#0,r6
GenerateDonutSuperStrip_stripVtx:

		mul.l	r8,r6
		sts	macl,r0
		mul.l	r9,r7
		sts	macl,r1
		add	r1,r0
		or	r12,r0
		mov.l	r0,@r4
		add	#4,r4

		mov	r7,r0
		add	#1,r0
		cmp/eq	r0,r11
		bf	GenerateDonutSuperStrip_nWrap1
		mov	#0,r0
GenerateDonutSuperStrip_nWrap1:

		mul.l	r8,r6
		sts	macl,r1
		mul.l	r9,r0
		sts	macl,r0
		add	r1,r0
		or	r12,r0
		mov.l	r0,@r4
		add	#4,r4

		add	#1,r6
		cmp/hs	r10,r6
		bf	GenerateDonutSuperStrip_stripVtx

		mul.l	r9,r7
		sts	macl,r0
		or	r12,r0
		mov.l	r0,@r4
		add	#4,r4

		mov	r7,r0
		add	#1,r0
		cmp/eq	r0,r11
		bf	GenerateDonutSuperStrip_nWrap2
		mov	#0,r0
GenerateDonutSuperStrip_nWrap2:

		mul.l	r9,r0
		sts	macl,r0
		or	r13,r0
		mov.l	r0,@r4
		add	#4,r4

		add	#1,r7
		cmp/hs	r11,r7
		bf	GenerateDonutSuperStrip_strip

		mov.l	@r15+,r8
		mov.l	@r15+,r9
		mov.l	@r15+,r10
		mov.l	@r15+,r11
		mov.l	@r15+,r12
		mov.l	@r15+,r13
		rts
		mov.l	@r15+,r14


	

				.ALIGN	2
GenerateDonut_DonutSteps1:	.LONG	DONUTSTEPS1
GenerateDonut_DonutSteps2:	.LONG	DONUTSTEPS2
GenerateDonut_DonutRadius1:	.SINGLE	1.0
GenerateDonut_DonutRadius2:	.SINGLE	2.0
GenerateDonut_65536:		.SINGLE	65536.0

!----------------------------------------------------------------------
! void CalculateDonutMorphVertices(Vtx* startvertex, Vtx* endvertex,
!	Vtx* startnormal, Vtx* endnormal,
!	MorphVtx* dest,  int count)
!
! Calculates morph-vertex data from two objects
!
! in	r4		Vtx*		Object 0 vertices
!	r5		Vtx*		Object 1 vertices
!	r6		Vtx*		Object 0 vertexnormals
!	r7		Vtx*		Object 1 vertexnormals
!	@(0,r15)	MorphVtx*	Output morph-vertices
!	@(4,r15)	int		Number of vertices

		.ALIGN	5
CalculateDonutMorphVertices:
		mov.l	r14,@-r15
		mov.l	r13,@-r15
		mov.l	r12,@-r15
		mov.l	r11,@-r15
		mov.l	r10,@-r15
		mov.l	r9,@-r15
		mov.l	r8,@-r15

		mov.l	@(7*4,r15),r8
		mov.l	@(8*4,r15),r9

CalculateDonutMorphVertices_vtx:
		fmov.s	@r6,fr0
		mov.l	@r6+,r0
		fmov.s	@r6,fr1
		mov.l	@r6+,r1
		fmov.s	@r6,fr2
		mov.l	@r6+,r2

		add	#4,r6
	
		fmov.s	@r7+,fr4
		fmov.s	@r7+,fr5
		fmov.s	@r7+,fr6

		add	#4,r7

		mov.l	r0,@(0,r8)
		mov.l	r1,@(4,r8)
		mov.l	r2,@(8,r8)

		fsub	fr0,fr4
		fsub	fr1,fr5
		fsub	fr2,fr6
		add	#28,r8

		fmov.s	fr6,@-r8
		fmov.s	fr5,@-r8
		fmov.s	fr4,@-r8
		add	#16,r8

		fmov.s	@r4,fr0
		mov.l	@r4+,r0
		fmov.s	@r4,fr1
		mov.l	@r4+,r1
		fmov.s	@r4,fr2
		mov.l	@r4+,r2

		add	#4,r4
	
		fmov.s	@r5+,fr4
		fmov.s	@r5+,fr5
		fmov.s	@r5+,fr6

		add	#4,r5

		mov.l	r0,@(0,r8)
		mov.l	r1,@(4,r8)
		mov.l	r2,@(8,r8)

		fsub	fr0,fr4
		fsub	fr1,fr5
		fsub	fr2,fr6
		add	#28,r8

		fmov.s	fr6,@-r8
		fmov.s	fr5,@-r8
		fmov.s	fr4,@-r8
		add	#16,r8

		dt	r9
		bf	CalculateDonutMorphVertices_vtx

		mov.l	@r15+,r8
		mov.l	@r15+,r9
		mov.l	@r15+,r10
		mov.l	@r15+,r11
		mov.l	@r15+,r12
		mov.l	@r15+,r13
		rts
		mov.l	@r15+,r14



!----------------------------------------------------------------------
! void DrawFrame(void)
!
! Draws one frame of the donut
!

		.ALIGN	5
DrawFrame:
		mov.l	r9,@-r15
		mov.l	r8,@-r15
		sts.l	pr,@-r15

		mov.l	DrawFrame_rdBeginScene,r0
		jsr	@r0
		mov	#0,r4				! Normal mode, please

		mov.l	DrawFrame_rdSetupDirectPath,r0
		jsr	@r0
		mov	#RENDER_LIST_OPAQUE,r4		! Opaque list as Direct Path

		mov.l	DrawFrame_OpaqueListHeader,r0	! Set opaque list header
		mov	#RENDER_LIST_OPAQUE,r4
		mov	r0,r5
		mov.l	DrawFrame_rdDrawPrimitives,r0
		jsr	@r0
		mov	#32,r6

		bsr	TransformAndDrawObject
		nop

		bsr	SetBorderGreen
		nop

		mov.l	DrawFrame_rdEndScene,r0		! Done with scene
		jsr	@r0
		nop

		mov.l	DrawFrame_rdFlip,r0
		jsr	@r0
		nop

		lds.l	@r15+,pr
		mov.l	@r15+,r8
		rts
		mov.l	@r15+,r9


				.ALIGN	5
DrawFrame_rdBeginScene:		.long	_rdBeginScene
DrawFrame_rdEndScene:		.long	_rdEndScene
DrawFrame_rdFlip:		.long	_rdFlip
DrawFrame_rdSetupDirectPath:	.long	_rdSetupDirectPath
DrawFrame_rdFinishDirectPath:	.long	_rdFinishDirectPath
DrawFrame_rdDrawPrimitives:	.long	_rdDrawPrimitives
DrawFrame_OpaqueListHeader:	.long	OpaqueListHeader



				.ALIGN	5
OpaqueListHeader:
				.long	0x80840002
				.long	0x90800000
				.long	0x20800440
				.long	0x00000000
				.long	-1,-1,-1,-1


!----------------------------------------------------------------------
! void TransformAndDrawObject(void)
!
! Transform donut and send it to TA
!

TransformAndDrawObject:
		sts.l	pr,@-r15

! Build transformation matrix
	
		mov.l	TransformAndDrawObject_XMat4x4_setRotation,r1
		mova	Animate_CurrentRotAngles,r0
		jsr	@r1
		mov	r0,r4

		mov.l	TransformAndDrawObject_XMat4x4_applyTranslation,r1
		mova	Animate_CurrentTranslation,r0
		jsr	@r1
		mov	r0,r4

		mov.l	TransformAndDrawObject_XMat4x4_store,r1
		mova	TransformAndDrawObject_TempMatrix,r0
		jsr	@r1
		mov	r0,r4

		frchg
		fldi0	fr3
		fldi0	fr7
		fldi0	fr11
		fldi0	fr12
		fldi0	fr13
		fldi0	fr14
		fldi1	fr15
		fneg	fr2
		fneg	fr6
		fneg	fr10
		frchg
		
! Morph: interpolate the morph-vertices to get an intermediate set of vertices, and do lighting

		mova	Animate_CurrentMorphPos,r0
		fmov.s	@r0,fr0
		ftrc	fr0,fpul
		fsca	fpul,dr0
		fldi1	fr2
		fldi1	fr3
		fadd	fr2,fr2
		fdiv	fr2,fr3
		fmov	fr3,fr4
		fmul	fr3,fr0
		fadd	fr0,fr4
		mov.l	TransformAndDrawObject_DestVertices,r4
		mov.l	TransformAndDrawObject_SourceMorphVertices,r5
		mov.l	TransformAndDrawObject_NumSourceVertices,r6
		bsr	InterpolateMorphVerticesAndLight
		nop

! Build transformation + projection matrix
	
		mov.l	TransformAndDrawObject_XMat4x4_load,r1
		mova	TransformAndDrawObject_TempMatrix,r0
		jsr	@r1
		mov	r0,r4

		mov.l	TransformAndDrawObject_XMat4x4_apply2DProjectionShearShuffle,r1
		mova	Animate_CurrentXYScale,r0
		mov	r0,r4
		mova	Animate_CurrentXYCenter,r0
		jsr	@r1
		mov	r0,r5

! Transform+draw the morphed object

		mov.l	TransformAndDrawObject_VertexBuffers,r0
		add	#RENDER_LIST_OPAQUE*32,r0
		mov.l	@(VertexBuffer_CurrentWritePtr,r0),r6
		mov.l	TransformAndDrawObject_QACR,r1
		mov.l	@(VertexBuffer_CurrentWriteArea,r0),r2
		mov.l	r2,@(0,r1)
		mov.l	r2,@(4,r1)
		mov.l	TransformAndDrawObject_SourceSuperStrip,r4
		mov.l	TransformAndDrawObject_DestVertices,r5

		bsr	DrawSuperStrip_gouraud
		nop

		mov	r0,r4
	
		mov.l	TransformAndDrawObject_VertexBuffers,r0
		add	#RENDER_LIST_OPAQUE*32,r0
		mov.l	r4,@(VertexBuffer_CurrentWritePtr,r0)

! Increase rotation angles

		mova	Animate_CurrentRotAngles,r0
		fmov.s	@r0+,fr0
		fmov.s	@r0+,fr1
		fmov.s	@r0+,fr2
		mova	Animate_CurrentRotSteps,r0
		fmov.s	@r0+,fr3
		fmov.s	@r0+,fr4
		fmov.s	@r0+,fr5
		mova	Animate_CurrentRotAngles+3*4,r0
		fadd	fr5,fr2
		fadd	fr4,fr1
		fadd	fr3,fr0
		fmov.s	fr2,@-r0
		fmov.s	fr1,@-r0
		fmov.s	fr0,@-r0

		mova	Animate_CurrentMorphPos,r0
		fmov.s	@r0+,fr0
		fmov.s	@r0,fr1
		fadd	fr1,fr0
		fmov.s	fr0,@-r0

		lds.l	@r15+,pr
		rts
		nop


								.ALIGN	4
TransformAndDrawObject_VertexBuffers:				.LONG	_VertexBuffers
TransformAndDrawObject_DestVertices:				.LONG	DonutVertices
TransformAndDrawObject_SourceMorphVertices:			.LONG	DonutMorphVertices
TransformAndDrawObject_NumSourceVertices:			.LONG	DONUTSTEPS1*DONUTSTEPS2
TransformAndDrawObject_SourceSuperStrip:			.LONG	DonutSuperStrip
TransformAndDrawObject_QACR:					.LONG	0xff000038
TransformAndDrawObject_XMat4x4_load:				.LONG	_XMat4x4_load
TransformAndDrawObject_XMat4x4_store:				.LONG	_XMat4x4_store
TransformAndDrawObject_XMat4x4_setIdentity:			.LONG	_XMat4x4_setIdentity
TransformAndDrawObject_XMat4x4_setRotation:			.LONG	_XMat4x4_setRotation
TransformAndDrawObject_XMat4x4_applyScale:			.LONG	_XMat4x4_applyScale
TransformAndDrawObject_XMat4x4_applyTranslation:		.LONG	_XMat4x4_applyTranslation
TransformAndDrawObject_XMat4x4_applyMatrix:			.LONG	_XMat4x4_applyMatrix
TransformAndDrawObject_XMat4x4_apply2DProjectionShearShuffle:	.LONG	_XMat4x4_apply2DProjectionShearShuffle
TransformAndDrawObject_ApplyLighting:				.LONG	ApplyLighting

	
				.ALIGN	4
Animate_CurrentRotAngles:	.float	0.0, 0.0, 0.0, 0.0
Animate_CurrentRotSteps:	.float	0.02, 0.012, 0.025, 0.0
Animate_CurrentTranslation:	.float	0.0, 0.0, 6.0, 0.0
Animate_CurrentXYScale:		.float	300.0, 300.0
Animate_CurrentXYCenter:	.float	320.0, 240.0
Animate_CurrentMorphPos:	.float	0.0
Animate_CurrentMorphStep:	.float	200.0
	
				.ALIGN	5
TransformAndDrawObject_TempMatrix:	.SPACE	64
TransformAndDrawObject_LightMatrix:
				.single	0.0, 0.0, 0.0, 0.0
				.single	0.0, 0.0, 0.0, 0.0
				.single	-1.0, -1.0, -1.0, -1.0
				.single	0.0, 0.0, 0.0, 0.0



!----------------------------------------------------------------------
! void DrawSuperStrip_gouraud(void* superStrip, void* vertices,
!	void* outputQueue, Vec4 lightVec)
!
! Transforms & draws a SuperStrip. Data is output in the following
! format: (code, x, y, rhw, -, -, argb, -)
!
! in	r4	void*	superstrip
!	r5	void*	vertices
!	r6	void*	output queue
!	fv4	Vec4	light vector
!	xmtrx	Mat4x4	vertex transformation
! out	r0	void*	current output queue pos
!
		.ALIGN	5
DrawSuperStrip_gouraud:
		fmov.s	fr15,@-r15
		fmov.s	fr14,@-r15
		fmov.s	fr13,@-r15
		fmov.s	fr12,@-r15
		mov.l	r14,@-r15
		mov.l	r13,@-r15
		mov.l	r12,@-r15
		mov.l	r11,@-r15
		mov.l	r10,@-r15
		mov.l	r9,@-r15
		mov.l	r8,@-r15

		fschg

		mov	#0x04,r7
		shll16	r7
		shll8	r7
		add	#-1,r7

		mov.l	@r4+,r9
DrawSuperStrip_gouraud_strip:

		mov.l	r9,@-r15

		mov.l	@r4+,r8
		add	#-32,r6

		mov.l	@r4+,r3
		dt	r8

		mov	r3,r2
		and	r7,r2

		add	r5,r2

		pref	@r2

		mov	r3,r0
		mov.l	@r4+,r3

		dt	r8

		fmov.d	@r2+,dr0



		fmov.d	@r2,dr2
		mov	r2,r11

		fldi1	fr3
		mov	r3,r2

		ftrv	xmtrx,fv0

		mov.l	@(4,r11),r1
		and	r7,r2

		add	r5,r2

		pref	@r2

		fmov	dr2,dr6
		fdiv	fr0,fr7

		mov	r0,r9

		mov	r3,r0
		mov.l	@r4+,r3

		fmov	dr0,dr4
		dt	r8

		bt/s	DrawSuperStrip_gouraud_singleStripVertex
	
		fmov.d	@r2+,dr0


DrawSuperStrip_gouraud_stripVertex:

		fmov.d	@r2,dr2
		mov	r2,r11

		fldi1	fr3
		mov	r3,r2

		mov.l	r1,@(32+24,r6)
		ftrv	xmtrx,fv0

		mov.l	@(4,r11),r1
		and	r7,r2

		add	r5,r2

		pref	@r2
		add	#32+16,r6

		fmov	dr6,dr8
		fmul	fr7,fr8

		!mov.l	r?,@(??,r6) ! Texture
		fmul	fr7,fr5

		fmov	dr2,dr6
		fdiv	fr0,fr7

		fmov.d	dr8,@-r6
		
		fmov.d	dr4,@-r6

		mov.l	r9,@r6
		mov	r0,r9

		mov	r3,r0
		mov.l	@r4+,r3

		fmov	dr0,dr4
		dt	r8

		pref	@r6
		bf/s	DrawSuperStrip_gouraud_stripVertex

		fmov.d	@r2+,dr0


DrawSuperStrip_gouraud_singleStripVertex:

		fmov.d	@r2,dr2
		mov	r2,r11

		fldi1	fr3
		mov	r3,r2

		mov.l	r1,@(32+24,r6)
		ftrv	xmtrx,fv0

		mov.l	@(4,r11),r1
		and	r7,r2

		add	r5,r2

		pref	@r2
		add	#32+16,r6

		fmov	dr6,dr8
		fmul	fr7,fr8

		!mov.l	r?,@(??,r6) ! Texture
		fmul	fr7,fr5

		fmov	dr2,dr6
		fdiv	fr0,fr7

		fmov.d	dr8,@-r6
		
		fmov.d	dr4,@-r6

		mov.l	r9,@r6
		mov	r0,r9

		mov	r3,r0

		fmov	dr0,dr4

		pref	@r6

		fmov.d	@r2+,dr0




		fmov.d	@r2,dr2
		mov	r2,r11

		fldi1	fr3

		mov.l	r1,@(32+24,r6)
		ftrv	xmtrx,fv0

		mov.l	@(4,r11),r1

		add	#32+16,r6

		fmov	dr6,dr8
		fmul	fr7,fr8

		!mov.l	r?,@(??,r6) ! Texture
		fmul	fr7,fr5

		fmov	dr2,dr6
		fdiv	fr0,fr7

		fmov.d	dr8,@-r6
		
		fmov.d	dr4,@-r6

		mov.l	r9,@r6
		mov	r0,r9

		fmov	dr0,dr4

		pref	@r6



		mov.l	r1,@(32+24,r6)

		add	#32+16,r6

		fmov	dr6,dr8
		fmul	fr7,fr8

		!mov.l	r?,@(??,r6) ! Texture
		fmul	fr7,fr5

		fmov.d	dr8,@-r6
		
		fmov.d	dr4,@-r6

		mov.l	r9,@r6

		pref	@r6

		add	#32,r6
	

		mov.l	@r15+,r9

		dt	r9
		bf	DrawSuperStrip_gouraud_strip

		fschg

		mov	r6,r0

		mov.l	@r15+,r8
		mov.l	@r15+,r9
		mov.l	@r15+,r10
		mov.l	@r15+,r11
		mov.l	@r15+,r12
		mov.l	@r15+,r13
		mov.l	@r15+,r14
		fmov.s	@r15+,fr12
		fmov.s	@r15+,fr13
		fmov.s	@r15+,fr14
		rts
		fmov.s	@r15+,fr15



!-----------------------------------------------------------------------------
! void InterpolateMorphVertices(Vtx* vertices, MorphVtx* morphVertices,
!		int numVertices, float scale)
!
! Interpolate morph vertices, outputting a morphed version of the object; also do (simple) diffuse lighting
!
! in	r4	Vtx*		Output vertices
!	r5	MorphVtx*	Input morph-vertices
!	r6	int		Number of vertices
!	fr4	float		Interpolation position


		.ALIGN	5
InterpolateMorphVerticesAndLight:
		fmov.s	fr15,@-r15
		fmov.s	fr14,@-r15
		fmov.s	fr13,@-r15
		fmov.s	fr12,@-r15

		shlr	r6
		add	#1,r6

		mova	InterpolateMorphVerticesAndLight_LightScale,r0

		frchg

		fmov.s	@r0,fr3

		fmul	fr3,fr0
		fmul	fr3,fr1
		fmul	fr3,fr2
		fmul	fr3,fr4
		fmul	fr3,fr5
		fmul	fr3,fr6
		fmul	fr3,fr7
		fmul	fr3,fr8
		fmul	fr3,fr9

		fldi0	fr3

		frchg

		fmov	fr4,fr0
	
		mov	r4,r0
		shlr16	r0
		mov.l	InterpolateMorphVerticesAndLight_QACR,r1
		shlr8	r0

		mov.l	@r1,r2
		mov.l	@(4,r1),r3
		mov.l	r0,@r1
		mov.l	r0,@(4,r1)
		mov.l	r3,@-r15
		mov.l	r2,@-r15
	
		mov.l	InterpolateMorphVerticesAndLight_SQAnd,r7
		mov.l	InterpolateMorphVerticesAndLight_SQOr,r1
		and	r4,r7
		or	r1,r7

		fschg

		fmov.d	@r5+,dr4
		fmov.d	@r5+,dr6
		fmov.d	@r5+,dr8
		fmov.d	@r5+,dr10

		fmov.d	@r5+,dr12
		fmac	fr0,fr8,fr4
		fmov.d	@r5+,dr14
		fmac	fr0,fr9,fr5
		fmov.d	@r5+,dr8
		fmac	fr0,fr10,fr6
		fmov.d	@r5+,dr10

		add	#32,r5
		pref	@r5

		fldi0	fr7
		add	#32,r5

		ftrv	xmtrx,fv4

		fmac	fr0,fr8,fr12
		fmac	fr0,fr9,fr13
		fmac	fr0,fr10,fr14

		ftrc	fr6,fpul

		pref	@r5

		add	#16,r7

InterpolateMorphVerticesAndLight_vtx:

		sts	fpul,r0
		add	#-(32+32),r5
		fmov.d	@r5+,dr4
		add	#32,r0
		fmov.d	@r5+,dr6
		cmp/pl	r0
		fmov.d	@r5+,dr8
		movt	r1
		fmov.d	@r5+,dr10
		neg	r1,r1

		ocbi	@r4

		fmov.d	dr14,@-r7
		and	r1,r0
		fmov.d	dr12,@-r7

		mov	r0,r1
		shll8	r1
		or	r1,r0
		shll8	r1
		or	r1,r0

		fmov.d	@r5+,dr12
		fmac	fr0,fr8,fr4
		fmov.d	@r5+,dr14
		fmac	fr0,fr9,fr5
		fmov.d	@r5+,dr8
		fmac	fr0,fr10,fr6
		fmov.d	@r5,dr10
		add	#32+8,r5
		pref	@r5

		fldi0	fr7
		add	#32,r5

		ftrv	xmtrx,fv4

		mov.l	r0,@(12,r7)
		add	#16+16,r7

		fmac	fr0,fr8,fr12
		fmac	fr0,fr9,fr13
		fmac	fr0,fr10,fr14

		ftrc	fr6,fpul

		pref	@r5

		sts	fpul,r0
		add	#-(32+32),r5
		fmov.d	@r5+,dr4
		add	#32,r0
		fmov.d	@r5+,dr6
		cmp/pl	r0
		fmov.d	@r5+,dr8
		movt	r1
		fmov.d	@r5+,dr10
		neg	r1,r1

		fmov.d	dr14,@-r7
		and	r1,r0
		fmov.d	dr12,@-r7

		mov	r0,r1
		shll8	r1
		or	r1,r0
		shll8	r1
		or	r1,r0

		mov.l	r0,@(12,r7)
		add	#32,r4

		pref	@r7
		add	#16+16,r7

		fmov.d	@r5+,dr12
		fmac	fr0,fr8,fr4
		fmov.d	@r5+,dr14
		fmac	fr0,fr9,fr5
		fmov.d	@r5+,dr8
		fmac	fr0,fr10,fr6
		fmov.d	@r5,dr10
		add	#32+8,r5
		pref	@r5

		fldi0	fr7
		add	#32,r5

		ftrv	xmtrx,fv4

		fmac	fr0,fr8,fr12
		dt	r6
		fmac	fr0,fr9,fr13
		fmac	fr0,fr10,fr14

		ftrc	fr6,fpul

		bf/s	InterpolateMorphVerticesAndLight_vtx
		pref	@r5

		mov	#0xff,r1
		shll16	r1
		shll8	r1
		add	#0x38,r1

		mov.l	@r15+,r2
		mov.l	@r15+,r3
		mov.l	r2,@r1
		mov.l	r3,@(4,r1)

		fschg

		fmov.s	@r15+,fr12
		fmov.s	@r15+,fr13
		fmov.s	@r15+,fr14
		rts
		fmov.s	@r15+,fr15

						.ALIGN	4
InterpolateMorphVerticesAndLight_LightScale:	.SINGLE	192.0
InterpolateMorphVerticesAndLight_QACR:		.LONG	0xff000038
InterpolateMorphVerticesAndLight_SQAnd:		.LONG	0x03ffffff
InterpolateMorphVerticesAndLight_SQOr:		.LONG	0xe0000000

